import os
import sys

import salt.modules.pip as pip
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PipTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {pip: {"__salt__": {"cmd.which_bin": lambda _: "pip"}}}

    def test__pip_bin_env(self):
        ret = pip._pip_bin_env(None, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
        if salt.utils.platform.is_windows():
            self.assertEqual(ret, "C:/Users/ch44d/Documents/salt/tests")
        else:
            self.assertIsNone(ret)

    def test__pip_bin_env_no_change(self):
        cwd = "C:/Users/ch44d/Desktop"
        ret = pip._pip_bin_env(cwd, "C:/Users/ch44d/Documents/salt/tests/pip.exe")
        self.assertEqual(ret, cwd)

    def test__pip_bin_env_no_bin_env(self):
        ret = pip._pip_bin_env(None, None)
        self.assertIsNone(ret)

    def test_install_frozen_app(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch("sys.frozen", True, create=True):
            with patch("sys._MEIPASS", True, create=True):
                with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                    pip.install(pkg)
                    expected = [
                        sys.executable,
                        "pip",
                        "install",
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        python_shell=False,
                        saltenv="base",
                        use_vt=False,
                        runas=None,
                    )

    def test_install_source_app(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch("sys.frozen", False, create=True):
            with patch("sys._MEIPASS", False, create=True):
                with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                    pip.install(pkg)
                    expected = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        python_shell=False,
                        saltenv="base",
                        use_vt=False,
                        runas=None,
                    )

    def test_fix4361(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(requirements="requirements.txt")
            expected_cmd = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--requirement",
                "requirements.txt",
            ]
            mock.assert_called_with(
                expected_cmd,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_editable_without_egg_fails(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                editable="git+https://github.com/saltstack/salt-testing.git",
            )

    def test_install_multiple_editable(self):
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        expected = [sys.executable, "-m", "pip", "install"]
        for item in editables:
            expected.extend(["--editable", item])

        # Passing editables as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(editable=editables)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing editables as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(editable=",".join(editables))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_multiple_pkgs_and_editables(self):
        pkgs = ["pep8", "salt"]
        editables = [
            "git+https://github.com/jek/blinker.git#egg=Blinker",
            "git+https://github.com/saltstack/salt-testing.git#egg=SaltTesting",
        ]

        expected = [sys.executable, "-m", "pip", "install"]
        expected.extend(pkgs)
        for item in editables:
            expected.extend(["--editable", item])

        # Passing editables as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=pkgs, editable=editables)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing editables as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=",".join(pkgs), editable=",".join(editables))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # As single string (just use the first element from pkgs and editables)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkgs=pkgs[0], editable=editables[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                pkgs[0],
                "--editable",
                editables[0],
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_issue5940_install_multiple_pip_mirrors(self):
        """
        test multiple pip mirrors.  This test only works with pip < 7.0.0
        """
        with patch.object(pip, "version", MagicMock(return_value="1.4")):
            mirrors = [
                "http://g.pypi.python.org",
                "http://c.pypi.python.org",
                "http://pypi.crate.io",
            ]

            expected = [sys.executable, "-m", "pip", "install", "--use-mirrors"]
            for item in mirrors:
                expected.extend(["--mirrors", item])
            expected.append("pep8")

            # Passing mirrors as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=mirrors)
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing mirrors as a comma separated list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=",".join(mirrors))
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--use-mirrors",
                "--mirrors",
                mirrors[0],
                "pep8",
            ]

            # As single string (just use the first element from mirrors)
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkgs=["pep8"], mirrors=mirrors[0])
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_with_multiple_find_links(self):
        find_links = [
            "http://g.pypi.python.org",
            "http://c.pypi.python.org",
            "http://pypi.crate.io",
        ]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in find_links:
            expected.extend(["--find-links", item])
        expected.append(pkg)

        # Passing mirrors as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=",".join(find_links))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Valid protos work?
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        expected = [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--find-links",
            find_links[0],
            pkg,
        ]

        # As single string (just use the first element from find_links)
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, find_links=find_links[0])
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Invalid proto raises exception
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                "'" + pkg + "'",
                find_links="sftp://pypi.crate.io",
            )

    def test_install_no_index_with_index_url_or_extra_index_url_raises(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                no_index=True,
                index_url="http://foo.tld",
            )

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                pip.install,
                no_index=True,
                extra_index_url="http://foo.tld",
            )

    def test_install_failed_cached_requirements(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            get_cached_requirements.return_value = False
            ret = pip.install(requirements="salt://my_test_reqs")
            self.assertEqual(False, ret["result"])
            self.assertIn("my_test_reqs", ret["comment"])

    def test_install_cached_requirements_used(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            get_cached_requirements.return_value = "my_cached_reqs"
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements="salt://requirements.txt")
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--requirement",
                    "my_cached_reqs",
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_venv(self):
        with patch("os.path") as mock_path:

            def join(*args):
                return os.path.normpath(os.sep.join(args))

            mock_path.is_file.return_value = True
            mock_path.isdir.return_value = True
            mock_path.join = join

            if salt.utils.platform.is_windows():
                venv_path = "C:\\test_env"
                bin_path = os.path.join(venv_path, "python.exe")
            else:
                venv_path = "/test_env"
                bin_path = os.path.join(venv_path, "python")

            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            pip_bin = MagicMock(return_value=[bin_path, "-m", "pip"])

            with patch.dict(pip.__salt__, {"cmd.run_all": mock}), patch.object(
                pip, "_get_pip_bin", pip_bin
            ):
                pip.install("mock", bin_env=venv_path)
                mock.assert_called_with(
                    [bin_path, "-m", "pip", "install", "mock"],
                    env={"VIRTUAL_ENV": venv_path},
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_log_argument_in_resulting_command(self):
        with patch("os.access") as mock_path:
            pkg = "pep8"
            log_path = "/tmp/pip-install.log"
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg, log=log_path)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--log",
                    log_path,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_non_writeable_log(self):
        with patch("os.path") as mock_path:
            # Let's fake a non-writable log file
            pkg = "pep8"
            log_path = "/tmp/pip-install.log"
            mock_path.exists.side_effect = IOError("Fooo!")
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(IOError, pip.install, pkg, log=log_path)

    def test_install_timeout_argument_in_resulting_command(self):
        # Passing an int
        pkg = "pep8"
        expected = [sys.executable, "-m", "pip", "install", "--timeout"]
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, timeout=10)
            mock.assert_called_with(
                expected + [10, pkg],
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing an int as a string
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, timeout="10")
            mock.assert_called_with(
                expected + ["10", pkg],
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing a non-int to timeout
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(ValueError, pip.install, pkg, timeout="a")

    def test_install_index_url_argument_in_resulting_command(self):
        pkg = "pep8"
        index_url = "http://foo.tld"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, index_url=index_url)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--index-url",
                index_url,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_extra_index_url_argument_in_resulting_command(self):
        pkg = "pep8"
        extra_index_url = "http://foo.tld"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, extra_index_url=extra_index_url)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--extra-index-url",
                extra_index_url,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_no_index_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_index=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-index", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_build_argument_in_resulting_command(self):
        pkg = "pep8"
        build = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, build=build)
            expected = [sys.executable, "-m", "pip", "install", "--build", build, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_target_argument_in_resulting_command(self):
        pkg = "pep8"
        target = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, target=target)
            expected = [sys.executable, "-m", "pip", "install", "--target", target, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_download_argument_in_resulting_command(self):
        pkg = "pep8"
        download = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, download=download)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--download",
                download,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_no_download_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_download=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-download", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_download_cache_dir_arguments_in_resulting_command(self):
        pkg = "pep8"
        cache_dir_arg_mapping = {
            "1.5.6": "--download-cache",
            "6.0": "--cache-dir",
        }
        download_cache = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            for pip_version, cmd_arg in cache_dir_arg_mapping.items():
                with patch(
                    "salt.modules.pip.version", MagicMock(return_value=pip_version)
                ):
                    # test `download_cache` kwarg
                    pip.install(pkg, download_cache="/tmp/foo")
                    expected = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        cmd_arg,
                        download_cache,
                        pkg,
                    ]
                    mock.assert_called_with(
                        expected,
                        saltenv="base",
                        runas=None,
                        use_vt=False,
                        python_shell=False,
                    )

                    # test `cache_dir` kwarg
                    pip.install(pkg, cache_dir="/tmp/foo")
                    mock.assert_called_with(
                        expected,
                        saltenv="base",
                        runas=None,
                        use_vt=False,
                        python_shell=False,
                    )

    def test_install_source_argument_in_resulting_command(self):
        pkg = "pep8"
        source = "/tmp/foo"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, source=source)
            expected = [sys.executable, "-m", "pip", "install", "--source", source, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_exists_action_argument_in_resulting_command(self):
        pkg = "pep8"
        for action in ("s", "i", "w", "b"):
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(pkg, exists_action=action)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--exists-action",
                    action,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

        # Test for invalid action
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError, pip.install, pkg, exists_action="d"
            )

    def test_install_install_options_argument_in_resulting_command(self):
        install_options = ["--exec-prefix=/foo/bar", "--install-scripts=/foo/bar/bin"]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in install_options:
            expected.extend(["--install-option", item])
        expected.append(pkg)

        # Passing options as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=install_options)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=",".join(install_options))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a single string entry
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, install_options=install_options[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--install-option",
                install_options[0],
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_global_options_argument_in_resulting_command(self):
        global_options = ["--quiet", "--no-user-cfg"]
        pkg = "pep8"

        expected = [sys.executable, "-m", "pip", "install"]
        for item in global_options:
            expected.extend(["--global-option", item])
        expected.append(pkg)

        # Passing options as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=global_options)
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=",".join(global_options))
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing mirrors as a single string entry
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, global_options=global_options[0])
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--global-option",
                global_options[0],
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_upgrade_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, upgrade=True)
            expected = [sys.executable, "-m", "pip", "install", "--upgrade", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_force_reinstall_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, force_reinstall=True)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--force-reinstall",
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_ignore_installed_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, ignore_installed=True)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--ignore-installed",
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_no_deps_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_deps=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-deps", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_no_install_argument_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, no_install=True)
            expected = [sys.executable, "-m", "pip", "install", "--no-install", pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_proxy_argument_in_resulting_command(self):
        pkg = "pep8"
        proxy = "salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(pkg, proxy=proxy)
            expected = [sys.executable, "-m", "pip", "install", "--proxy", proxy, pkg]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_proxy_false_argument_in_resulting_command(self):
        """
        Checking that there is no proxy set if proxy arg is set to False
        even if the global proxy is set.
        """
        pkg = "pep8"
        proxy = False
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.install(pkg, proxy=proxy)
                expected = [sys.executable, "-m", "pip", "install", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_global_proxy_in_resulting_command(self):
        """
        Checking that there is proxy set if global proxy is set.
        """
        pkg = "pep8"
        proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.install(pkg)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--proxy",
                    proxy,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_multiple_requirements_arguments_in_resulting_command(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
            get_cached_requirements.side_effect = cached_reqs
            requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

            expected = [sys.executable, "-m", "pip", "install"]
            for item in cached_reqs:
                expected.extend(["--requirement", item])

            # Passing option as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=requirements)
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a comma separated list
            get_cached_requirements.side_effect = cached_reqs
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=",".join(requirements))
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a single string entry
            get_cached_requirements.side_effect = [cached_reqs[0]]
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.install(requirements=requirements[0])
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--requirement",
                    cached_reqs[0],
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_install_extra_args_arguments_in_resulting_command(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.install(
                pkg, extra_args=[{"--latest-pip-kwarg": "param"}, "--latest-pip-arg"]
            )
            expected = [
                sys.executable,
                "-m",
                "pip",
                "install",
                pkg,
                "--latest-pip-kwarg",
                "param",
                "--latest-pip-arg",
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

    def test_install_extra_args_arguments_recursion_error(self):
        pkg = "pep8"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):

            self.assertRaises(
                TypeError,
                lambda: pip.install(
                    pkg, extra_args=[{"--latest-pip-kwarg": ["param1", "param2"]}]
                ),
            )

            self.assertRaises(
                TypeError,
                lambda: pip.install(
                    pkg, extra_args=[{"--latest-pip-kwarg": [{"--too-deep": dict()}]}]
                ),
            )

    def test_uninstall_multiple_requirements_arguments_in_resulting_command(self):
        with patch(
            "salt.modules.pip._get_cached_requirements"
        ) as get_cached_requirements:
            cached_reqs = ["my_cached_reqs-1", "my_cached_reqs-2"]
            get_cached_requirements.side_effect = cached_reqs
            requirements = ["salt://requirements-1.txt", "salt://requirements-2.txt"]

            expected = [sys.executable, "-m", "pip", "uninstall", "-y"]
            for item in cached_reqs:
                expected.extend(["--requirement", item])

            # Passing option as a list
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=requirements)
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a comma separated list
            get_cached_requirements.side_effect = cached_reqs
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=",".join(requirements))
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

            # Passing option as a single string entry
            get_cached_requirements.side_effect = [cached_reqs[0]]
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                pip.uninstall(requirements=requirements[0])
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    "--requirement",
                    cached_reqs[0],
                ]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_global_proxy_in_resulting_command(self):
        """
        Checking that there is proxy set if global proxy is set.
        """
        pkg = "pep8"
        proxy = "http://salt-user:salt-passwd@salt-proxy:3128"
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.uninstall(pkg)
                expected = [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    "--proxy",
                    proxy,
                    pkg,
                ]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_proxy_false_argument_in_resulting_command(self):
        """
        Checking that there is no proxy set if proxy arg is set to False
        even if the global proxy is set.
        """
        pkg = "pep8"
        proxy = False
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        config_mock = {
            "proxy_host": "salt-proxy",
            "proxy_port": "3128",
            "proxy_username": "salt-user",
            "proxy_password": "salt-passwd",
        }
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch.dict(pip.__opts__, config_mock):
                pip.uninstall(pkg, proxy=proxy)
                expected = [sys.executable, "-m", "pip", "uninstall", "-y", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_uninstall_log_argument_in_resulting_command(self):
        pkg = "pep8"
        log_path = "/tmp/pip-install.log"

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, log=log_path)
            expected = [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "-y",
                "--log",
                log_path,
                pkg,
            ]
            mock.assert_called_with(
                expected,
                saltenv="base",
                cwd=None,
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Let's fake a non-writable log file
        with patch("os.path") as mock_path:
            mock_path.exists.side_effect = IOError("Fooo!")
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(IOError, pip.uninstall, pkg, log=log_path)

    def test_uninstall_timeout_argument_in_resulting_command(self):
        pkg = "pep8"
        expected = [sys.executable, "-m", "pip", "uninstall", "-y", "--timeout"]
        # Passing an int
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, timeout=10)
            mock.assert_called_with(
                expected + [10, pkg],
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing an int as a string
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            pip.uninstall(pkg, timeout="10")
            mock.assert_called_with(
                expected + ["10", pkg],
                cwd=None,
                saltenv="base",
                runas=None,
                use_vt=False,
                python_shell=False,
            )

        # Passing a non-int to timeout
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(ValueError, pip.uninstall, pkg, timeout="a")

    def test_freeze_command(self):
        expected = [sys.executable, "-m", "pip", "freeze"]
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.freeze()
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )
                self.assertEqual(ret, eggs)

        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        # Passing env_vars passes them to underlying command?
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.freeze(env_vars={"foo": "bar"})
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                    env={"foo": "bar"},
                )
                self.assertEqual(ret, eggs)

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError,
                    pip.freeze,
                )

    def test_freeze_command_with_all(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pip==0.9.1",
            "pycrypto==2.6",
            "setuptools==20.10.1",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
                ret = pip.freeze()
                expected = [sys.executable, "-m", "pip", "freeze", "--all"]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )
                self.assertEqual(ret, eggs)

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.0.1")):
                self.assertRaises(
                    CommandExecutionError,
                    pip.freeze,
                )

    def test_list_command(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock_version = "6.1.1"
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch(
                "salt.modules.pip.version", MagicMock(return_value=mock_version)
            ):
                ret = pip.list_()
                expected = [sys.executable, "-m", "pip", "freeze"]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertEqual(
                    ret,
                    {
                        "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                        "M2Crypto": "0.21.1",
                        "bbfreeze-loader": "1.1.0",
                        "bbfreeze": "1.1.0",
                        "pip": mock_version,
                        "pycrypto": "2.6",
                    },
                )

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError,
                    pip.list_,
                )

    def test_list_command_with_all(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pip==9.0.1",
            "pycrypto==2.6",
            "setuptools==20.10.1",
        ]
        # N.B.: this is deliberately different from the "output" of pip freeze.
        # This is to demonstrate that the version reported comes from freeze
        # instead of from the pip.version function.
        mock_version = "9.0.0"
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch(
                "salt.modules.pip.version", MagicMock(return_value=mock_version)
            ):
                ret = pip.list_()
                expected = [sys.executable, "-m", "pip", "freeze", "--all"]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertEqual(
                    ret,
                    {
                        "SaltTesting-dev": "git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8",
                        "M2Crypto": "0.21.1",
                        "bbfreeze-loader": "1.1.0",
                        "bbfreeze": "1.1.0",
                        "pip": "9.0.1",
                        "pycrypto": "2.6",
                        "setuptools": "20.10.1",
                    },
                )

        # Non zero returncode raises exception?
        mock = MagicMock(return_value={"retcode": 1, "stderr": "CABOOOOMMM!"})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                self.assertRaises(
                    CommandExecutionError,
                    pip.list_,
                )

    def test_list_command_with_prefix(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.list_(prefix="bb")
                expected = [sys.executable, "-m", "pip", "freeze"]
                mock.assert_called_with(
                    expected,
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertEqual(ret, {"bbfreeze-loader": "1.1.0", "bbfreeze": "1.1.0"})

    def test_list_upgrades_legacy(self):
        eggs = [
            "apache-libcloud (Current: 1.1.0 Latest: 2.2.1 [wheel])",
            "appdirs (Current: 1.4.1 Latest: 1.4.3 [wheel])",
            "awscli (Current: 1.11.63 Latest: 1.12.1 [sdist])",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.list_upgrades()
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "list", "--outdated"],
                    cwd=None,
                    runas=None,
                )
                self.assertEqual(
                    ret,
                    {
                        "apache-libcloud": "2.2.1 [wheel]",
                        "appdirs": "1.4.3 [wheel]",
                        "awscli": "1.12.1 [sdist]",
                    },
                )

    def test_list_upgrades_gt9(self):
        eggs = """[{"latest_filetype": "wheel", "version": "1.1.0", "name": "apache-libcloud", "latest_version": "2.2.1"},
                {"latest_filetype": "wheel", "version": "1.4.1", "name": "appdirs", "latest_version": "1.4.3"},
                {"latest_filetype": "sdist", "version": "1.11.63", "name": "awscli", "latest_version": "1.12.1"}
                ]"""
        mock = MagicMock(return_value={"retcode": 0, "stdout": "{}".format(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="9.1.1")):
                ret = pip.list_upgrades()
                mock.assert_called_with(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "list",
                        "--outdated",
                        "--format=json",
                    ],
                    cwd=None,
                    runas=None,
                )
                self.assertEqual(
                    ret,
                    {
                        "apache-libcloud": "2.2.1 [wheel]",
                        "appdirs": "1.4.3 [wheel]",
                        "awscli": "1.12.1 [sdist]",
                    },
                )

    def test_is_installed_true(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.is_installed(pkgname="bbfreeze")
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "freeze"],
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertTrue(ret)

    def test_is_installed_false(self):
        eggs = [
            "M2Crypto==0.21.1",
            "-e git+git@github.com:s0undt3ch/salt-testing.git@9ed81aa2f918d59d3706e56b18f0782d1ea43bf8#egg=SaltTesting-dev",
            "bbfreeze==1.1.0",
            "bbfreeze-loader==1.1.0",
            "pycrypto==2.6",
        ]
        mock = MagicMock(return_value={"retcode": 0, "stdout": "\n".join(eggs)})
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="6.1.1")):
                ret = pip.is_installed(pkgname="notexist")
                mock.assert_called_with(
                    [sys.executable, "-m", "pip", "freeze"],
                    cwd=None,
                    runas=None,
                    python_shell=False,
                    use_vt=False,
                )
                self.assertFalse(ret)

    def test_install_pre_argument_in_resulting_command(self):
        pkg = "pep8"
        # Lower than 1.4 versions don't end up with `--pre` in the resulting output
        mock = MagicMock(
            side_effect=[
                {"retcode": 0, "stdout": "pip 1.2.0 /path/to/site-packages/pip"},
                {"retcode": 0, "stdout": ""},
            ]
        )
        with patch.dict(pip.__salt__, {"cmd.run_all": mock}):
            with patch("salt.modules.pip.version", MagicMock(return_value="1.3")):
                pip.install(pkg, pre_releases=True)
                expected = [sys.executable, "-m", "pip", "install", pkg]
                mock.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

        mock_run = MagicMock(return_value="pip 1.4.1 /path/to/site-packages/pip")
        mock_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(
            pip.__salt__, {"cmd.run_stdout": mock_run, "cmd.run_all": mock_run_all}
        ):
            with patch(
                "salt.modules.pip._get_pip_bin", MagicMock(return_value=["pip"])
            ):
                pip.install(pkg, pre_releases=True)
                expected = ["pip", "install", "--pre", pkg]
                mock_run_all.assert_called_with(
                    expected,
                    saltenv="base",
                    runas=None,
                    use_vt=False,
                    python_shell=False,
                )

    def test_resolve_requirements_chain_function(self):
        """ensure requirements chain can handle special cases"""

        class FakeFopen:
            def __init__(self, filename):
                d = {
                    "requirements-0.txt": (
                        b"--index-url http://fake.com/simple\n\n"
                        b"one  # -r wrong.txt, other\n"
                        b"two # --requirement wrong.exe;some\n"
                        b"three\n"
                        b"-r requirements-1.txt\n"
                        b"# nothing\n"
                    ),
                    "requirements-1.txt": (
                        "four\n"
                        "five\n"
                        "--requirement=requirements-2.txt\t# --requirements-2.txt\n\n"
                    ),
                    "requirements-2.txt": b"""six""",
                    "requirements-3.txt": (
                        b"# some comment\n"
                        b"-e git+ssh://git.example.com/MyProject#egg=MyProject # the project\n"
                        b"seven\n"
                        b"-e git+ssh://git.example.com/Example#egg=example\n"
                        b"eight # -e something#or other\n"
                        b"--requirement requirements-4.txt\n\n"
                    ),
                    "requirements-4.txt": "",
                }
                self.val = d[filename]

            def __enter__(self):
                return self

            def __exit__(self, *args, **kwargs):
                pass

            def read(self):
                return self.val

        with patch("salt.utils.files.fopen", FakeFopen):
            chain = pip._resolve_requirements_chain(
                ["requirements-0.txt", "requirements-3.txt"]
            )
        self.assertEqual(
            chain,
            [
                "requirements-0.txt",
                "requirements-1.txt",
                "requirements-2.txt",
                "requirements-3.txt",
                "requirements-4.txt",
            ],
        )

    # TODO: When we switch to pytest, mark parametrized with None for user as well -W. Werner, 2020-06-23
    def test_when_upgrade_is_called_and_there_are_available_upgrades_it_should_call_correct_command(
        self,
    ):
        fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
        expected_user = "fnord"
        with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
            "salt.modules.pip.list_upgrades", autospec=True, return_value=["fnord"]
        ), patch(
            "salt.modules.pip._get_pip_bin",
            autospec=True,
            return_value=["some-other-pip"],
        ):
            pip.upgrade(user=expected_user)

            fake_run_all.assert_any_call(
                ["some-other-pip", "install", "-U", "freeze", "--all", "fnord"],
                runas=expected_user,
                cwd=None,
                use_vt=False,
            )

    # TODO: Yeah, parametrize the user here -W. Werner, 2020-07-07
    def test_when_list_upgrades_is_provided_a_user_it_should_be_passed_to_the_version_command(
        self,
    ):
        fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": "{}"})
        expected_user = "fnord"

        def all_new_commands(*args, **kwargs):
            """
            Without this, mutating the return value mutates the return value
            for EVERYTHING.
            """
            return ["some-other-pip"]

        with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
            "salt.modules.pip._get_pip_bin",
            autospec=True,
            side_effect=all_new_commands,
        ):
            pip._clear_context()
            pip.list_upgrades(user=expected_user)
            fake_run_all.assert_any_call(
                ["some-other-pip", "--version"],
                runas=expected_user,
                cwd=None,
                python_shell=False,
            )

    # TODO: Yeah, parametrize the user here -W. Werner, 2020-07-07
    def test_when_install_is_provided_a_user_it_should_be_passed_to_the_version_command(
        self,
    ):
        fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": "{}"})
        expected_user = "fnord"

        def all_new_commands(*args, **kwargs):
            """
            Without this, mutating the return value mutates the return value
            for EVERYTHING.
            """
            return ["some-other-pip"]

        with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
            "salt.modules.pip._get_pip_bin",
            autospec=True,
            side_effect=all_new_commands,
        ):
            pip._clear_context()
            pip.install(user=expected_user)
            fake_run_all.assert_any_call(
                ["some-other-pip", "--version"],
                runas=expected_user,
                cwd=None,
                python_shell=False,
            )

    # TODO: When we switch to pytest, parameterize the user - None and fnord should be enough, maybe a 3rd user -W. Werner, 2020-07-07
    def test_when_version_is_called_with_a_user_it_should_be_passed_to_undelying_runas(
        self,
    ):
        fake_run_all = MagicMock(return_value={"retcode": 0, "stdout": ""})
        expected_user = "fnord"
        with patch.dict(pip.__salt__, {"cmd.run_all": fake_run_all}), patch(
            "salt.modules.pip.list_upgrades", autospec=True, return_value=["fnord"]
        ), patch(
            "salt.modules.pip._get_pip_bin",
            autospec=True,
            return_value=["some-new-pip"],
        ):
            pip.version(user=expected_user)
            fake_run_all.assert_called_with(
                ["some-new-pip", "--version"],
                runas=expected_user,
                cwd=None,
                python_shell=False,
            )



"""
    :codeauthor: :email:`Vernon Cole <vernondcole@gmail.com>`
"""


import os

import salt.utils.sdb as sdb
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class SdbTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.sdb
    """

    @classmethod
    def setUpClass(cls):
        cls.TEMP_DATABASE_FILE = os.path.join(RUNTIME_VARS.TMP, "test_sdb.sqlite")
        cls.sdb_opts = {
            "extension_modules": "",
            "optimization_order": [0, 1, 2],
            "test_sdb_data": {
                "driver": "sqlite3",
                "database": cls.TEMP_DATABASE_FILE,
                "table": "sdb",
                "create_table": True,
            },
        }

    @classmethod
    def tearDownClass(cls):
        try:
            os.unlink(cls.TEMP_DATABASE_FILE)
        except OSError:
            pass

    def setup_loader_modules(self):
        return {sdb: {}}

    # test with SQLite database key not presest

    def test_sqlite_get_not_found(self):
        what = sdb.sdb_get("sdb://test_sdb_data/thisKeyDoesNotExist", self.sdb_opts)
        self.assertEqual(what, None)

    # test with SQLite database write and read

    def test_sqlite_get_found(self):
        expected = {b"name": b"testone", b"number": 46}
        sdb.sdb_set("sdb://test_sdb_data/test1", expected, self.sdb_opts)
        resp = sdb.sdb_get("sdb://test_sdb_data/test1", self.sdb_opts)
        self.assertEqual(resp, expected)



"""
    :codeauthor: Ted Strzalkowski (tedski@gmail.com)


    tests.unit.modules.mdadm_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import re
import sys

import pytest
import salt.modules.mdadm_raid as mdadm
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MdadmTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {mdadm: {}}

    @pytest.mark.skipif(
        sys.version_info < (3, 6), reason="Py3.5 dictionaries are not ordered"
    )
    def test_create(self):
        mock = MagicMock(return_value="salt")
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}), patch(
            "salt.utils.path.which", lambda exe: exe
        ):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                test_mode=False,
                force=True,
                chunk=256,
            )
            self.assertEqual("salt", ret)
            mock.assert_called_with(
                [
                    "mdadm",
                    "-C",
                    "/dev/md0",
                    "-R",
                    "-v",
                    "-l",
                    "5",
                    "--force",
                    "--chunk",
                    "256",
                    "-e",
                    "default",
                    "-n",
                    "3",
                    "/dev/sdb1",
                    "/dev/sdc1",
                    "/dev/sdd1",
                ],
                python_shell=False,
            )

    @pytest.mark.skipif(
        sys.version_info < (3, 6), reason="Py3.5 dictionaries are not ordered"
    )
    def test_create_metadata(self):
        mock = MagicMock(return_value="salt")
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}), patch(
            "salt.utils.path.which", lambda exe: exe
        ):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                metadata=0.9,
                test_mode=False,
                force=True,
                chunk=256,
            )
            self.assertEqual("salt", ret)
            mock.assert_called_with(
                [
                    "mdadm",
                    "-C",
                    "/dev/md0",
                    "-R",
                    "-v",
                    "-l",
                    "5",
                    "--force",
                    "--chunk",
                    "256",
                    "-e",
                    "0.9",
                    "-n",
                    "3",
                    "/dev/sdb1",
                    "/dev/sdc1",
                    "/dev/sdd1",
                ],
                python_shell=False,
            )

    def test_create_test_mode(self):
        mock = MagicMock()
        with patch.dict(mdadm.__salt__, {"cmd.run": mock}):
            ret = mdadm.create(
                "/dev/md0",
                5,
                devices=["/dev/sdb1", "/dev/sdc1", "/dev/sdd1"],
                force=True,
                chunk=256,
                test_mode=True,
            )
            self.assertEqual(
                sorted(
                    "mdadm -C /dev/md0 -R -v --chunk 256 "
                    "--force -l 5 -e default -n 3 "
                    "/dev/sdb1 /dev/sdc1 /dev/sdd1".split()
                ),
                sorted(ret.split()),
            )
            assert not mock.called, "test mode failed, cmd.run called"

    def test_examine(self):
        """
        Test for mdadm_raid.examine
        """
        mock = MagicMock(
            return_value=(
                "ARRAY /dev/md/pool metadata=1.2"
                " UUID=567da122:fb8e445e:55b853e0:81bd0a3e name=positron:pool"
            )
        )
        with patch.dict(mdadm.__salt__, {"cmd.run_stdout": mock}):
            self.assertEqual(
                mdadm.examine("/dev/md0"),
                {
                    "ARRAY /dev/md/pool metadata": (
                        "1.2 UUID=567da122:fb8e445e:55b853e0:81bd0a3e"
                        " name=positron:pool"
                    )
                },
            )
            mock.assert_called_with(
                "mdadm -Y -E /dev/md0", ignore_retcode=False, python_shell=False
            )

    def test_examine_quiet(self):
        """
        Test for mdadm_raid.examine
        """
        mock = MagicMock(return_value="")
        with patch.dict(mdadm.__salt__, {"cmd.run_stdout": mock}):
            self.assertEqual(mdadm.examine("/dev/md0", quiet=True), {})
            mock.assert_called_with(
                "mdadm -Y -E /dev/md0", ignore_retcode=True, python_shell=False
            )

    def test_device_match_regex_pattern(self):
        assert re.match(
            mdadm._VOL_REGEX_PATTERN_MATCH.format("/dev/md/1"),
            "ARRAY /dev/md/1  metadata=1.2 UUID=51f245bc:a1402c8a:2d598e79:589c07cf"
            " name=tst-ob-001:1",
        )
        assert not re.match(
            mdadm._VOL_REGEX_PATTERN_MATCH.format("/dev/md/1"),
            "ARRAY /dev/md/10  metadata=1.2 UUID=51f245bc:a1402c8a:2d598e79:589c07cf"
            " name=tst-ob-001:1",
        )



##
# Copyright 2018 Pure Storage Inc
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

Management of Pure Storage FlashBlade

Installation Prerequisites
--------------------------
- You will need the ``purity_fb`` python package in your python installation
  path that is running salt.

  .. code-block:: bash

      pip install purity_fb

- Configure Pure Storage FlashBlade authentication. Use one of the following
  three methods.

  1) From the minion config

  .. code-block:: yaml

        pure_tags:
          fb:
            san_ip: management vip or hostname for the FlashBlade
            api_token: A valid api token for the FlashBlade being managed

  2) From environment (PUREFB_IP and PUREFB_API)
  3) From the pillar (PUREFB_IP and PUREFB_API)

:maintainer: Simon Dodsley (simon@purestorage.com)
:maturity: new
:requires: purity_fb
:platform: all

.. versionadded:: 2019.2.0

"""


import os
from datetime import datetime

from salt.exceptions import CommandExecutionError

try:
    from purity_fb import PurityFb, FileSystem, FileSystemSnapshot, SnapshotSuffix
    from purity_fb import rest, NfsRule, ProtocolRule

    HAS_PURITY_FB = True
except ImportError:
    HAS_PURITY_FB = False

__docformat__ = "restructuredtext en"

__virtualname__ = "purefb"


def __virtual__():
    """
    Determine whether or not to load this module
    """
    if HAS_PURITY_FB:
        return __virtualname__
    return (
        False,
        "purefb execution module not loaded: purity_fb python library not available.",
    )


def _get_blade():
    """
    Get Pure Storage FlasBlade configuration

    1) From the minion config
        pure_tags:
          fb:
            san_ip: management vip or hostname for the FlashBlade
            api_token: A valid api token for the FlashBlade being managed
    2) From environment (PUREFB_IP and PUREFB_API)
    3) From the pillar (PUREFB_IP and PUREFB_API)

    """

    try:
        blade_name = __opts__["pure_tags"]["fb"].get("san_ip")
        api_token = __opts__["pure_tags"]["fb"].get("api_token")
        if blade_name and api:
            blade = PurityFb(blade_name)
            blade.disable_verify_ssl()
    except (KeyError, NameError, TypeError):
        try:
            blade_name = os.environ.get("PUREFB_IP")
            api_token = os.environ.get("PUREFB_API")
            if blade_name:
                blade = PurityFb(blade_name)
                blade.disable_verify_ssl()
        except (ValueError, KeyError, NameError):
            try:
                api_token = __pillar__["PUREFB_API"]
                blade = PurityFb(__pillar__["PUREFB_IP"])
                blade.disable_verify_ssl()
            except (KeyError, NameError):
                raise CommandExecutionError(
                    "No Pure Storage FlashBlade credentials found."
                )
    try:
        blade.login(api_token)
    except Exception:  # pylint: disable=broad-except
        raise CommandExecutionError("Pure Storage FlashBlade authentication failed.")
    return blade


def _get_fs(name, blade):
    """
    Private function to
    check for existence of a filesystem
    """
    _fs = []
    _fs.append(name)
    try:
        res = blade.file_systems.list_file_systems(names=_fs)
        return res.items[0]
    except rest.ApiException:
        return None


def _get_snapshot(name, suffix, blade):
    """
    Return name of Snapshot
    or None
    """
    try:
        filt = "source='{}' and suffix='{}'".format(name, suffix)
        res = blade.file_system_snapshots.list_file_system_snapshots(filter=filt)
        return res.items[0]
    except rest.ApiException:
        return None


def _get_deleted_fs(name, blade):
    """
    Private function to check
    if a file systeem has already been deleted
    """
    try:
        _fs = _get_fs(name, blade)
        if _fs and _fs.destroyed:
            return _fs
    except rest.ApiException:
        return None


def snap_create(name, suffix=None):
    """

    Create a filesystem snapshot on a Pure Storage FlashBlade.

    Will return False if filesystem selected to snap does not exist.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem to snapshot
    suffix : string
        if specificed forces snapshot name suffix. If not specified defaults to timestamp.

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.snap_create foo
        salt '*' purefb.snap_create foo suffix=bar

    """
    blade = _get_blade()
    if suffix is None:
        suffix = "snap-" + str(
            (datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0, 0)).total_seconds()
        )
        suffix = suffix.replace(".", "")
    if _get_fs(name, blade) is not None:
        try:
            source = []
            source.append(name)
            blade.file_system_snapshots.create_file_system_snapshots(
                sources=source, suffix=SnapshotSuffix(suffix)
            )
            return True
        except rest.ApiException:
            return False
    else:
        return False


def snap_delete(name, suffix=None, eradicate=False):
    """

    Delete a filesystem snapshot on a Pure Storage FlashBlade.

    Will return False if selected snapshot does not exist.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem
    suffix : string
        name of snapshot
    eradicate : boolean
        Eradicate snapshot after deletion if True. Default is False

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.snap_delete foo suffix=snap eradicate=True

    """
    blade = _get_blade()
    if _get_snapshot(name, suffix, blade) is not None:
        try:
            snapname = name + "." + suffix
            new_attr = FileSystemSnapshot(destroyed=True)
            blade.file_system_snapshots.update_file_system_snapshots(
                name=snapname, attributes=new_attr
            )
        except rest.ApiException:
            return False
        if eradicate is True:
            try:
                blade.file_system_snapshots.delete_file_system_snapshots(name=snapname)
                return True
            except rest.ApiException:
                return False
        else:
            return True
    else:
        return False


def snap_eradicate(name, suffix=None):
    """

    Eradicate a deleted filesystem snapshot on a Pure Storage FlashBlade.

    Will return False if snapshot is not in a deleted state.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem
    suffix : string
        name of snapshot

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.snap_eradicate foo suffix=snap

    """
    blade = _get_blade()
    if _get_snapshot(name, suffix, blade) is not None:
        snapname = name + "." + suffix
        try:
            blade.file_system_snapshots.delete_file_system_snapshots(name=snapname)
            return True
        except rest.ApiException:
            return False
    else:
        return False


def fs_create(
    name, size=None, proto="NFS", nfs_rules="*(rw,no_root_squash)", snapshot=False
):
    """

    Create a filesystem on a Pure Storage FlashBlade.

    Will return False if filesystem already exists.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem (truncated to 63 characters)
    proto : string
        (Optional) Sharing protocol (NFS, CIFS or HTTP). If not specified default is NFS
    snapshot: boolean
        (Optional) Are snapshots enabled on the filesystem. Default is False
    nfs_rules : string
        (Optional) export rules for NFS. If not specified default is
        ``*(rw,no_root_squash)``. Refer to Pure Storage documentation for
        formatting rules.
    size : string
        if specified capacity of filesystem. If not specified default to 32G.
        Refer to Pure Storage documentation for formatting rules.

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.fs_create foo proto=CIFS
        salt '*' purefb.fs_create foo size=10T

    """
    if len(name) > 63:
        name = name[0:63]
    blade = _get_blade()
    print(proto)
    if _get_fs(name, blade) is None:
        if size is None:
            size = __utils__["stringutils.human_to_bytes"]("32G")
        else:
            size = __utils__["stringutils.human_to_bytes"](size)
        if proto.lower() == "nfs":
            fs_obj = FileSystem(
                name=name,
                provisioned=size,
                fast_remove_directory_enabled=True,
                snapshot_directory_enabled=snapshot,
                nfs=NfsRule(enabled=True, rules=nfs_rules),
            )
        elif proto.lower() == "cifs":
            fs_obj = FileSystem(
                name=name,
                provisioned=size,
                fast_remove_directory_enabled=True,
                snapshot_directory_enabled=snapshot,
                smb=ProtocolRule(enabled=True),
            )
        elif proto.lower() == "http":
            fs_obj = FileSystem(
                name=name,
                provisioned=size,
                fast_remove_directory_enabled=True,
                snapshot_directory_enabled=snapshot,
                http=ProtocolRule(enabled=True),
            )
        else:
            return False
        try:
            blade.file_systems.create_file_systems(fs_obj)
            return True
        except rest.ApiException:
            return False
    else:
        return False


def fs_delete(name, eradicate=False):
    """

    Delete a share on a Pure Storage FlashBlade.

    Will return False if filesystem doesn't exist or is already in a deleted state.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem
    eradicate : boolean
        (Optional) Eradicate filesystem after deletion if True. Default is False

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.fs_delete foo eradicate=True

    """
    blade = _get_blade()
    if _get_fs(name, blade) is not None:
        try:
            blade.file_systems.update_file_systems(
                name=name,
                attributes=FileSystem(
                    nfs=NfsRule(enabled=False),
                    smb=ProtocolRule(enabled=False),
                    http=ProtocolRule(enabled=False),
                    destroyed=True,
                ),
            )
        except rest.ApiException:
            return False
        if eradicate is True:
            try:
                blade.file_systems.delete_file_systems(name)
                return True
            except rest.ApiException:
                return False
        else:
            return True
    else:
        return False


def fs_eradicate(name):
    """

    Eradicate a deleted filesystem on a Pure Storage FlashBlade.

    Will return False is filesystem is not in a deleted state.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.fs_eradicate foo

    """
    blade = _get_blade()
    if _get_deleted_fs(name, blade) is not None:
        try:
            blade.file_systems.delete_file_systems(name)
            return True
        except rest.ApiException:
            return False
    else:
        return False


def fs_extend(name, size):
    """

    Resize an existing filesystem on a Pure Storage FlashBlade.

    Will return False if new size is less than or equal to existing size.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem
    size : string
        New capacity of filesystem.
        Refer to Pure Storage documentation for formatting rules.

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.fs_extend foo 10T

    """
    attr = {}
    blade = _get_blade()
    _fs = _get_fs(name, blade)
    if _fs is not None:
        if __utils__["stringutils.human_to_bytes"](size) > _fs.provisioned:
            try:
                attr["provisioned"] = __utils__["stringutils.human_to_bytes"](size)
                n_attr = FileSystem(**attr)
                blade.file_systems.update_file_systems(name=name, attributes=n_attr)
                return True
            except rest.ApiException:
                return False
        else:
            return False
    else:
        return False


def fs_update(name, rules, snapshot=False):
    """

    Update filesystem on a Pure Storage FlashBlade.

    Allows for change of NFS export rules and enabling/disabled
    of snapshotting capability.

    .. versionadded:: 2019.2.0

    name : string
        name of filesystem
    rules : string
        NFS export rules for filesystem
        Refer to Pure Storage documentation for formatting rules.
    snapshot: boolean
        (Optional) Enable/Disable snapshots on the filesystem. Default is False

    CLI Example:

    .. code-block:: bash

        salt '*' purefb.fs_nfs_update foo rules='10.234.112.23(ro), 10.234.112.24(rw)' snapshot=True

    """
    blade = _get_blade()
    attr = {}
    _fs = _get_fs(name, blade)
    if _fs is not None:
        try:
            if _fs.nfs.enabled:
                attr["nfs"] = NfsRule(rules=rules)
            attr["snapshot_directory_enabled"] = snapshot
            n_attr = FileSystem(**attr)
            blade.file_systems.update_file_systems(name=name, attributes=n_attr)
            return True
        except rest.ApiException:
            return False
    else:
        return False



"""
unit tests for the cache runner
"""

import os

import pytest
import salt.runners.queue as queue_mod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class QueueTest(TestCase, LoaderModuleMockMixin):
    """
    Validate the queue runner
    """

    def setup_loader_modules(self):
        return {
            queue_mod: {
                "__opts__": {
                    "sock_dir": os.path.join(RUNTIME_VARS.TMP, "queue-runner-sock-dir"),
                    "transport": "zeromq",
                }
            }
        }

    def test_insert_runner(self):
        queue_insert = MagicMock(return_value=True)
        with patch.object(queue_mod, "insert", queue_insert):
            queue_mod.insert_runner("test.stdout_print", queue="salt")
        expected_call = {
            "queue": "salt",
            "items": {"fun": "test.stdout_print", "args": [], "kwargs": {}},
            "backend": "pgjsonb",
        }
        queue_insert.assert_called_once_with(**expected_call)

    @pytest.mark.slow_test
    def test_process_runner(self):
        ret = [{"fun": "test.stdout_print", "args": [], "kwargs": {}}]

        queue_pop = MagicMock(return_value=ret)
        test_stdout_print = MagicMock(return_value=True)
        with patch.dict(queue_mod.__salt__, {"test.stdout_print": test_stdout_print}):
            with patch.object(queue_mod, "pop", queue_pop):
                queue_mod.process_runner(queue="salt")
            queue_pop.assert_called_once_with(
                is_runner=True, queue="salt", quantity=1, backend="pgjsonb"
            )
            test_stdout_print.assert_called_once_with()
            queue_pop.assert_called_once_with(
                is_runner=True, queue="salt", quantity=1, backend="pgjsonb"
            )



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import salt.modules.mac_group as mac_group
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

HAS_GRP = True
try:
    import grp
except ImportError:
    HAS_GRP = False


@skipIf(not HAS_GRP, "Missing required library 'grp'")
class MacGroupTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for the salt.modules.mac_group module
    """

    def setup_loader_modules(self):
        return {mac_group: {}}

    # 'add' function tests: 6

    def test_add_group_exists(self):
        """
        Tests if the group already exists or not
        """
        mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
        with patch("salt.modules.mac_group.info", MagicMock(return_value=mock_group)):
            self.assertRaises(CommandExecutionError, mac_group.add, "test")

    def test_add_whitespace(self):
        """
        Tests if there is whitespace in the group name
        """
        with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_group.add, "white space")

    def test_add_underscore(self):
        """
        Tests if the group name starts with an underscore or not
        """
        with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_group.add, "_Test")

    def test_add_gid_int(self):
        """
        Tests if the gid is an int or not
        """
        with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
            self.assertRaises(SaltInvocationError, mac_group.add, "foo", "foo")

    def test_add_gid_exists(self):
        """
        Tests if the gid is already in use or not
        """
        with patch("salt.modules.mac_group.info", MagicMock(return_value={})), patch(
            "salt.modules.mac_group._list_gids", MagicMock(return_value=["3456"])
        ):
            self.assertRaises(CommandExecutionError, mac_group.add, "foo", 3456)

    def test_add(self):
        """
        Tests if specified group was added
        """
        mock_ret = MagicMock(return_value=0)
        with patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
            "salt.modules.mac_group.info", MagicMock(return_value={})
        ), patch("salt.modules.mac_group._list_gids", MagicMock(return_value=[])):
            self.assertTrue(mac_group.add("test", 500))

    # 'delete' function tests: 4

    def test_delete_whitespace(self):
        """
        Tests if there is whitespace in the group name
        """
        self.assertRaises(SaltInvocationError, mac_group.delete, "white space")

    def test_delete_underscore(self):
        """
        Tests if the group name starts with an underscore or not
        """
        self.assertRaises(SaltInvocationError, mac_group.delete, "_Test")

    def test_delete_group_exists(self):
        """
        Tests if the group to be deleted exists or not
        """
        with patch("salt.modules.mac_group.info", MagicMock(return_value={})):
            self.assertTrue(mac_group.delete("test"))

    def test_delete(self):
        """
        Tests if the specified group was deleted
        """
        mock_ret = MagicMock(return_value=0)
        mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
        with patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
            "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
        ):
            self.assertTrue(mac_group.delete("test"))

    # 'info' function tests: 2

    def test_info_whitespace(self):
        """
        Tests if there is whitespace in the group name
        """
        self.assertRaises(SaltInvocationError, mac_group.info, "white space")

    def test_info(self):
        """
        Tests the return of group information
        """
        mock_getgrall = [grp.struct_group(("foo", "*", 20, ["test"]))]
        with patch("grp.getgrall", MagicMock(return_value=mock_getgrall)):
            ret = {"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}
            self.assertEqual(mac_group.info("foo"), ret)

    # '_format_info' function tests: 1

    def test_format_info(self):
        """
        Tests the formatting of returned group information
        """
        data = grp.struct_group(("wheel", "*", 0, ["root"]))
        ret = {"passwd": "*", "gid": 0, "name": "wheel", "members": ["root"]}
        self.assertEqual(mac_group._format_info(data), ret)

    # 'getent' function tests: 1

    def test_getent(self):
        """
        Tests the return of information on all groups
        """
        mock_getgrall = [grp.struct_group(("foo", "*", 20, ["test"]))]
        with patch("grp.getgrall", MagicMock(return_value=mock_getgrall)):
            ret = [{"passwd": "*", "gid": 20, "name": "foo", "members": ["test"]}]
            self.assertEqual(mac_group.getent(), ret)

    # 'chgid' function tests: 4

    def test_chgid_gid_int(self):
        """
        Tests if gid is an integer or not
        """
        self.assertRaises(SaltInvocationError, mac_group.chgid, "foo", "foo")

    def test_chgid_group_exists(self):
        """
        Tests if the group id exists or not
        """
        mock_pre_gid = MagicMock(return_value="")
        with patch.dict(mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}), patch(
            "salt.modules.mac_group.info", MagicMock(return_value={})
        ):
            self.assertRaises(CommandExecutionError, mac_group.chgid, "foo", 4376)

    def test_chgid_gid_same(self):
        """
        Tests if the group id is the same as argument
        """
        mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
        mock_pre_gid = MagicMock(return_value=0)
        with patch.dict(mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}), patch(
            "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
        ):
            self.assertTrue(mac_group.chgid("test", 0))

    def test_chgid(self):
        """
        Tests the gid for a named group was changed
        """
        mock_group = {"passwd": "*", "gid": 0, "name": "test", "members": ["root"]}
        mock_pre_gid = MagicMock(return_value=0)
        mock_ret = MagicMock(return_value=0)
        with patch.dict(
            mac_group.__salt__, {"file.group_to_gid": mock_pre_gid}
        ), patch.dict(mac_group.__salt__, {"cmd.retcode": mock_ret}), patch(
            "salt.modules.mac_group.info", MagicMock(return_value=mock_group)
        ):
            self.assertTrue(mac_group.chgid("test", 500))



"""
integration tests for mac_ports
"""

import pytest
from tests.support.case import ModuleCase
from tests.support.helpers import runs_on


@pytest.mark.skip_if_not_root
@runs_on(kernel="Darwin")
@pytest.mark.skip_if_binaries_missing("port")
class MacPortsModuleTest(ModuleCase):
    """
    Validate the mac_ports module
    """

    AGREE_INSTALLED = False

    def setUp(self):
        """
        Get current settings
        """
        self.AGREE_INSTALLED = "agree" in self.run_function("pkg.list_pkgs")
        self.run_function("pkg.refresh_db")

    def tearDown(self):
        """
        Reset to original settings
        """
        if not self.AGREE_INSTALLED:
            self.run_function("pkg.remove", ["agree"])

    @pytest.mark.destructive_test
    def test_list_pkgs(self):
        """
        Test pkg.list_pkgs
        """
        self.run_function("pkg.install", ["agree"])
        self.assertIsInstance(self.run_function("pkg.list_pkgs"), dict)
        self.assertIn("agree", self.run_function("pkg.list_pkgs"))

    @pytest.mark.destructive_test
    def test_latest_version(self):
        """
        Test pkg.latest_version
        """
        self.run_function("pkg.install", ["agree"])
        result = self.run_function("pkg.latest_version", ["agree"], refresh=False)
        self.assertIsInstance(result, dict)
        self.assertIn("agree", result)

    @pytest.mark.destructive_test
    def test_remove(self):
        """
        Test pkg.remove
        """
        self.run_function("pkg.install", ["agree"])
        removed = self.run_function("pkg.remove", ["agree"])
        self.assertIsInstance(removed, dict)
        self.assertIn("agree", removed)

    @pytest.mark.destructive_test
    def test_install(self):
        """
        Test pkg.install
        """
        self.run_function("pkg.remove", ["agree"])
        installed = self.run_function("pkg.install", ["agree"])
        self.assertIsInstance(installed, dict)
        self.assertIn("agree", installed)

    def test_list_upgrades(self):
        """
        Test pkg.list_upgrades
        """
        self.assertIsInstance(
            self.run_function("pkg.list_upgrades", refresh=False), dict
        )

    @pytest.mark.destructive_test
    def test_upgrade_available(self):
        """
        Test pkg.upgrade_available
        """
        self.run_function("pkg.install", ["agree"])
        self.assertFalse(
            self.run_function("pkg.upgrade_available", ["agree"], refresh=False)
        )

    def test_refresh_db(self):
        """
        Test pkg.refresh_db
        """
        self.assertTrue(self.run_function("pkg.refresh_db"))

    @pytest.mark.destructive_test
    def test_upgrade(self):
        """
        Test pkg.upgrade
        """
        results = self.run_function("pkg.upgrade", refresh=False)
        self.assertIsInstance(results, dict)
        self.assertTrue(results["result"])



"""
Tests the localfs tokens interface.
"""

import os

import salt.exceptions
import salt.tokens.localfs
import salt.utils.files
from tests.support.helpers import with_tempdir
from tests.support.mock import patch
from tests.support.unit import TestCase


class CalledWith:
    def __init__(self, func, called_with=None):
        self.func = func
        if called_with is None:
            self.called_with = []
        else:
            self.called_with = called_with

    def __call__(self, *args, **kwargs):
        self.called_with.append((args, kwargs))
        return self.func(*args, **kwargs)


class WriteTokenTest(TestCase):
    @with_tempdir()
    def test_write_token(self, tmpdir):
        """
        Validate tokens put in place with an atomic move
        """
        opts = {"token_dir": tmpdir}
        fopen = CalledWith(salt.utils.files.fopen)
        rename = CalledWith(os.rename)
        with patch("salt.utils.files.fopen", fopen), patch("os.rename", rename):
            tdata = salt.tokens.localfs.mk_token(opts, {})
        assert "token" in tdata
        t_path = os.path.join(tmpdir, tdata["token"])
        temp_t_path = "{}.tmp".format(t_path)
        assert len(fopen.called_with) == 1, len(fopen.called_with)
        assert fopen.called_with == [((temp_t_path, "w+b"), {})], fopen.called_with
        assert len(rename.called_with) == 1, len(rename.called_with)
        assert rename.called_with == [((temp_t_path, t_path), {})], rename.called_with


class TestLocalFS(TestCase):
    def setUp(self):
        # Default expected data
        self.expected_data = {"this": "is", "some": "token data"}

    @with_tempdir()
    def test_get_token_should_return_token_if_exists(self, tempdir):
        opts = {"token_dir": tempdir}
        tok = salt.tokens.localfs.mk_token(
            opts=opts,
            tdata=self.expected_data,
        )["token"]
        actual_data = salt.tokens.localfs.get_token(opts=opts, tok=tok)
        self.assertDictEqual(self.expected_data, actual_data)

    @with_tempdir()
    def test_get_token_should_raise_SaltDeserializationError_if_token_file_is_empty(
        self, tempdir
    ):
        opts = {"token_dir": tempdir}
        tok = salt.tokens.localfs.mk_token(
            opts=opts,
            tdata=self.expected_data,
        )["token"]
        with salt.utils.files.fopen(os.path.join(tempdir, tok), "w") as f:
            f.truncate()
        with self.assertRaises(salt.exceptions.SaltDeserializationError) as e:
            salt.tokens.localfs.get_token(opts=opts, tok=tok)

    @with_tempdir()
    def test_get_token_should_raise_SaltDeserializationError_if_token_file_is_malformed(
        self, tempdir
    ):
        opts = {"token_dir": tempdir}
        tok = salt.tokens.localfs.mk_token(
            opts=opts,
            tdata=self.expected_data,
        )["token"]
        with salt.utils.files.fopen(os.path.join(tempdir, tok), "w") as f:
            f.truncate()
            f.write("this is not valid msgpack data")
        with self.assertRaises(salt.exceptions.SaltDeserializationError) as e:
            salt.tokens.localfs.get_token(opts=opts, tok=tok)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.powerpath as powerpath
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PowerpathTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.powerpath
    """

    def setup_loader_modules(self):
        return {powerpath: {}}

    def test_has_powerpath(self):
        """
        Test for powerpath
        """
        with patch("os.path.exists") as mock_exists:
            mock_exists.return_value = True
            self.assertTrue(powerpath.has_powerpath())

            mock_exists.return_value = False
            self.assertFalse(powerpath.has_powerpath())

    def test_list_licenses(self):
        """
        Test to returns a list of applied powerpath license keys
        """
        with patch.dict(
            powerpath.__salt__, {"cmd.run": MagicMock(return_value="A\nB")}
        ):
            self.assertListEqual(powerpath.list_licenses(), [])

    def test_add_license(self):
        """
        Test to add a license
        """
        with patch.object(powerpath, "has_powerpath", return_value=False):
            self.assertDictEqual(
                powerpath.add_license("key"),
                {
                    "output": "PowerPath is not installed",
                    "result": False,
                    "retcode": -1,
                },
            )

        mock = MagicMock(return_value={"retcode": 1, "stderr": "stderr"})
        with patch.object(powerpath, "has_powerpath", return_value=True):
            with patch.dict(powerpath.__salt__, {"cmd.run_all": mock}):
                self.assertDictEqual(
                    powerpath.add_license("key"),
                    {"output": "stderr", "result": False, "retcode": 1},
                )

    def test_remove_license(self):
        """
        Test to remove a license
        """
        with patch.object(powerpath, "has_powerpath", return_value=False):
            self.assertDictEqual(
                powerpath.remove_license("key"),
                {
                    "output": "PowerPath is not installed",
                    "result": False,
                    "retcode": -1,
                },
            )

        mock = MagicMock(return_value={"retcode": 1, "stderr": "stderr"})
        with patch.object(powerpath, "has_powerpath", return_value=True):
            with patch.dict(powerpath.__salt__, {"cmd.run_all": mock}):
                self.assertDictEqual(
                    powerpath.remove_license("key"),
                    {"output": "stderr", "result": False, "retcode": 1},
                )



"""
:maintainer:    Alberto Planas <aplanas@suse.com>
:maturity:      new
:depends:       None
:platform:      Linux
"""

import logging

log = logging.getLogger(__name__)

__func_alias__ = {
    "filter_": "filter",
}


def _udev(udev_info, key):
    """
    Return the value for a udev key.

    The `key` parameter is a lower case text joined by dots. For
    example, 'e.id_bus' will represent the key for
    `udev_info['E']['ID_BUS']`.

    """
    k, _, r = key.partition(".")
    if not k:
        return udev_info
    if not isinstance(udev_info, dict):
        return "n/a"
    if not r:
        return udev_info.get(k.upper(), "n/a")
    return _udev(udev_info.get(k.upper(), {}), r)


def _match(udev_info, match_info):
    """
    Check if `udev_info` match the information from `match_info`.
    """
    res = True
    for key, value in match_info.items():
        udev_value = _udev(udev_info, key)
        if isinstance(udev_value, dict):
            # If is a dict we probably make a mistake in key from
            # match_info, as is not accessing a final value
            log.warning(
                "The key %s for the udev information dictionary is not a leaf element",
                key,
            )
            continue

        # Converting both values to sets make easy to see if there is
        # a coincidence between both values
        value = set(value) if isinstance(value, list) else {value}
        udev_value = set(udev_value) if isinstance(udev_value, list) else {udev_value}
        res = res and (value & udev_value)
    return res


def filter_(udev_in=None, udev_ex=None):
    """
    Returns a list of devices, filtered under udev keys.

    udev_in
        A dictionary of key:values that are expected in the device
        udev information

    udev_ex
        A dictionary of key:values that are not expected in the device
        udev information (excluded)

    The key is a lower case string, joined by dots, that represent a
    path in the udev information dictionary. For example, 'e.id_bus'
    will represent the udev entry `udev['E']['ID_BUS']`

    If the udev entry is a list, the algorithm will check that at
    least one item match one item of the value of the parameters.

    Returns list of devices that match `udev_in` and do not match
    `udev_ex`.

    CLI Example:

    .. code-block:: bash

       salt '*' devinfo.filter udev_in='{"e.id_bus": "ata"}'

    """

    udev_in = udev_in if udev_in else {}
    udev_ex = udev_ex if udev_ex else {}

    all_devices = __grains__["disks"]

    # Get the udev information only one time
    udev_info = {d: __salt__["udev.info"](d) for d in all_devices}

    devices_udev_key_in = {d for d in all_devices if _match(udev_info[d], udev_in)}
    devices_udev_key_ex = {
        d for d in all_devices if _match(udev_info[d], udev_ex) if udev_ex
    }

    return sorted(devices_udev_key_in - devices_udev_key_ex)


def _hwinfo_parse_short(report):
    """Parse the output of hwinfo and return a dictionary"""
    result = {}
    current_result = {}
    key_counter = 0
    for line in report.strip().splitlines():
        if line.startswith("    "):
            key = key_counter
            key_counter += 1
            current_result[key] = line.strip()
        elif line.startswith("  "):
            key, value = line.strip().split(" ", 1)
            current_result[key] = value.strip()
        elif line.endswith(":"):
            key = line[:-1]
            value = {}
            result[key] = value
            current_result = value
            key_counter = 0
        else:
            log.error("Error parsing hwinfo short output: %s", line)

    return result


def _hwinfo_parse_full(report):
    """Parse the output of hwinfo and return a dictionary"""
    result = {}
    result_stack = []
    level = 0
    for line in report.strip().splitlines():
        current_level = line.count("  ")
        if level != current_level or len(result_stack) != result_stack:
            result_stack = result_stack[:current_level]
            level = current_level
        line = line.strip()

        # Ignore empty lines
        if not line:
            continue

        # Initial line of a segment
        if level == 0:
            key, value = line.split(":", 1)
            sub_result = {}
            result[key] = sub_result
            # The first line contains also a sub-element
            key, value = value.strip().split(": ", 1)
            sub_result[key] = value
            result_stack.append(sub_result)
            level += 1
            continue

        # Line is a note
        if line.startswith("[") or ":" not in line:
            sub_result = result_stack[-1]
            sub_result["Note"] = line if not line.startswith("[") else line[1:-1]
            continue

        key, value = line.split(":", 1)
        key, value = key.strip(), value.strip()
        sub_result = result_stack[-1]
        # If there is a value and it not starts with hash, this is a
        # (key, value) entry. But there are exception on the rule,
        # like when is about 'El Torito info', that is the begining of
        # a new dictorionart.
        if value and not value.startswith("#") and key != "El Torito info":
            if key == "I/O Port":
                key = "I/O Ports"
            elif key == "Config Status":
                value = dict(item.split("=") for item in value.split(", "))
            elif key in ("Driver", "Driver Modules"):
                value = value.replace('"', "").split(", ")
            elif key in ("Tags", "Device Files", "Features"):
                # We cannot split by ', ', as using spaces in
                # inconsisten in some fields
                value = [v.strip() for v in value.split(",")]
            else:
                if value.startswith('"'):
                    value = value[1:-1]

            # If there is a collision, we store it as a list
            if key in sub_result:
                current_value = sub_result[key]
                if type(current_value) is not list:
                    current_value = [current_value]
                if value not in current_value:
                    current_value.append(value)
                if len(current_value) == 1:
                    value = current_value[0]
                else:
                    value = current_value
            sub_result[key] = value
        else:
            if value.startswith("#"):
                value = {"Handle": value}
            elif key == "El Torito info":
                value = value.split(", ")
                value = {
                    "platform": value[0].split()[-1],
                    "bootable": "no" if "not" in value[1] else "yes",
                }
            else:
                value = {}

            sub_result[key] = value
            result_stack.append(value)
            level += 1

    return result


def _hwinfo_parse(report, short):
    """Parse the output of hwinfo and return a dictionary"""
    if short:
        return _hwinfo_parse_short(report)
    else:
        return _hwinfo_parse_full(report)


def _hwinfo_efi():
    """Return information about EFI"""
    return {
        "efi": __grains__["efi"],
        "efi-secure-boot": __grains__["efi-secure-boot"],
    }


def _hwinfo_memory():
    """Return information about the memory"""
    return {
        "mem_total": __grains__["mem_total"],
    }


def _hwinfo_network(short):
    """Return network information"""
    info = {
        "fqdn": __grains__["fqdn"],
        "ip_interfaces": __grains__["ip_interfaces"],
    }

    if not short:
        info["dns"] = __grains__["dns"]

    return info


def hwinfo(items=None, short=True, listmd=False, devices=None):
    """
    Probe for hardware

    items
        List of hardware items to inspect. Default ['bios', 'cpu', 'disk',
        'memory', 'network', 'partition']

    short
        Show only a summary. Default True.

    listmd
        Report RAID devices. Default False.

    devices
        List of devices to show information from. Default None.

    CLI Example:

    .. code-block:: bash

       salt '*' devinfo.hwinfo
       salt '*' devinfo.hwinfo items='["disk"]' short=no
       salt '*' devinfo.hwinfo items='["disk"]' short=no devices='["/dev/sda"]'
       salt '*' devinfo.hwinfo devices=/dev/sda

    """
    result = {}

    if not items:
        items = ["bios", "cpu", "disk", "memory", "network", "partition"]
    if not isinstance(items, (list, tuple)):
        items = [items]

    if not devices:
        devices = []
    if devices and not isinstance(devices, (list, tuple)):
        devices = [devices]

    cmd = ["hwinfo"]
    for item in items:
        cmd.append("--{}".format(item))

    if short:
        cmd.append("--short")

    if listmd:
        cmd.append("--listmd")

    for device in devices:
        cmd.append("--only {}".format(device))

    out = __salt__["cmd.run_stdout"](cmd)
    result["hwinfo"] = _hwinfo_parse(out, short)

    if "bios" in items:
        result["bios grains"] = _hwinfo_efi()

    if "memory" in items:
        result["memory grains"] = _hwinfo_memory()

    if "network" in items:
        result["network grains"] = _hwinfo_network(short)

    return result



import pytest
import salt.utils.path
import salt.utils.platform
import salt.utils.systemd
from tests.support.case import ModuleCase


@pytest.mark.destructive_test
@pytest.mark.windows_whitelisted
class ServiceModuleTest(ModuleCase):
    """
    Module testing the service module
    """

    def setUp(self):
        self.service_name = "cron"
        cmd_name = "crontab"
        os_family = self.run_function("grains.get", ["os_family"])
        os_release = self.run_function("grains.get", ["osrelease"])
        if os_family == "RedHat":
            if os_release[0] == "7":
                self.skipTest(
                    "Disabled on CentOS 7 until we can fix SSH connection issues."
                )
            self.service_name = "crond"
        elif os_family == "Arch":
            self.service_name = "sshd"
            cmd_name = "systemctl"
        elif os_family == "NILinuxRT":
            self.service_name = "syslog"
            cmd_name = "syslog-ng"
        elif os_family == "MacOS":
            self.service_name = "org.ntp.ntpd"
            if int(os_release.split(".")[1]) >= 13:
                self.service_name = "com.apple.AirPlayXPCHelper"
        elif salt.utils.platform.is_windows():
            self.service_name = "Spooler"

        self.pre_srv_status = self.run_function("service.status", [self.service_name])
        self.pre_srv_enabled = (
            True
            if self.service_name in self.run_function("service.get_enabled")
            else False
        )

        if (
            salt.utils.path.which(cmd_name) is None
            and not salt.utils.platform.is_windows()
        ):
            self.skipTest("{} is not installed".format(cmd_name))

    def tearDown(self):
        post_srv_status = self.run_function("service.status", [self.service_name])
        post_srv_enabled = (
            True
            if self.service_name in self.run_function("service.get_enabled")
            else False
        )

        if post_srv_status != self.pre_srv_status:
            if self.pre_srv_status:
                self.run_function("service.enable", [self.service_name])
            else:
                self.run_function("service.disable", [self.service_name])

        if post_srv_enabled != self.pre_srv_enabled:
            if self.pre_srv_enabled:
                self.run_function("service.enable", [self.service_name])
            else:
                self.run_function("service.disable", [self.service_name])
        del self.service_name

    @pytest.mark.flaky(max_runs=4)
    @pytest.mark.slow_test
    def test_service_status_running(self):
        """
        test service.status execution module
        when service is running
        """
        self.run_function("service.start", [self.service_name])
        check_service = self.run_function("service.status", [self.service_name])
        self.assertTrue(check_service)

    @pytest.mark.slow_test
    def test_service_status_dead(self):
        """
        test service.status execution module
        when service is dead
        """
        self.run_function("service.stop", [self.service_name])
        check_service = self.run_function("service.status", [self.service_name])
        self.assertFalse(check_service)

    @pytest.mark.slow_test
    def test_service_restart(self):
        """
        test service.restart
        """
        self.assertTrue(self.run_function("service.restart", [self.service_name]))

    @pytest.mark.slow_test
    def test_service_enable(self):
        """
        test service.get_enabled and service.enable module
        """
        # disable service before test
        self.assertTrue(self.run_function("service.disable", [self.service_name]))

        self.assertTrue(self.run_function("service.enable", [self.service_name]))
        self.assertIn(self.service_name, self.run_function("service.get_enabled"))

    @pytest.mark.slow_test
    def test_service_disable(self):
        """
        test service.get_disabled and service.disable module
        """
        # enable service before test
        self.assertTrue(self.run_function("service.enable", [self.service_name]))

        self.assertTrue(self.run_function("service.disable", [self.service_name]))
        if salt.utils.platform.is_darwin():
            self.assertTrue(self.run_function("service.disabled", [self.service_name]))
        else:
            self.assertIn(self.service_name, self.run_function("service.get_disabled"))

    @pytest.mark.slow_test
    def test_service_disable_doesnot_exist(self):
        """
        test service.get_disabled and service.disable module
        when service name does not exist
        """
        # enable service before test
        srv_name = "doesnotexist"
        enable = self.run_function("service.enable", [srv_name])
        systemd = salt.utils.systemd.booted()

        # check service was not enabled
        try:
            self.assertFalse(enable)
        except AssertionError:
            self.assertIn("ERROR", enable)

        # check service was not disabled
        if (
            tuple(
                self.run_function("grains.item", ["osrelease_info"])["osrelease_info"]
            )
            == (14, 0o4)
            and not systemd
        ):
            # currently upstart does not have a mechanism to report if disabling a service fails if does not exist
            self.assertTrue(self.run_function("service.disable", [srv_name]))
        elif (
            self.run_function("grains.item", ["os"])["os"] == "Debian"
            and self.run_function("grains.item", ["osmajorrelease"])["osmajorrelease"]
            < 9
            and systemd
        ):
            # currently disabling a service via systemd that does not exist
            # on Debian 8 results in a True return code
            self.assertTrue(self.run_function("service.disable", [srv_name]))
        else:
            try:
                disable = self.run_function("service.disable", [srv_name])
                self.assertFalse(disable)
            except AssertionError:
                self.assertTrue("error" in disable.lower())

        if salt.utils.platform.is_darwin():
            self.assertEqual(
                self.run_function("service.disabled", [srv_name]),
                "ERROR: Service not found: {}".format(srv_name),
            )
        else:
            self.assertNotIn(srv_name, self.run_function("service.get_disabled"))

    @pytest.mark.skip_unless_on_windows
    @pytest.mark.slow_test
    def test_service_get_service_name(self):
        """
        test service.get_service_name
        """
        ret = self.run_function("service.get_service_name")
        self.assertIn(self.service_name, ret.values())



"""
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
"""

import pytest
import salt.states.beacon as beacon
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {beacon: {}}


def test_present():
    """
    Test to ensure a job is present in the beacon.
    """
    beacon_name = "ps"

    ret = {"name": beacon_name, "changes": {}, "result": False, "comment": ""}

    mock_mod = MagicMock(return_value=ret)
    mock_lst = MagicMock(side_effect=[{beacon_name: {}}, {beacon_name: {}}, {}, {}])
    with patch.dict(
        beacon.__salt__,
        {
            "beacons.list": mock_lst,
            "beacons.modify": mock_mod,
            "beacons.add": mock_mod,
        },
    ):
        assert beacon.present(beacon_name) == ret

        with patch.dict(beacon.__opts__, {"test": False}):
            assert beacon.present(beacon_name) == ret

        with patch.dict(beacon.__opts__, {"test": True}):
            ret.update({"result": True})
            assert beacon.present(beacon_name) == ret


def test_absent():
    """
    Test to ensure a job is absent from the schedule.
    """
    beacon_name = "ps"

    ret = {"name": beacon_name, "changes": {}, "result": False, "comment": ""}

    mock_mod = MagicMock(return_value=ret)
    mock_lst = MagicMock(side_effect=[{beacon_name: {}}, {}])
    with patch.dict(
        beacon.__salt__, {"beacons.list": mock_lst, "beacons.delete": mock_mod}
    ):
        with patch.dict(beacon.__opts__, {"test": False}):
            assert beacon.absent(beacon_name) == ret

        with patch.dict(beacon.__opts__, {"test": True}):
            comt = "ps not configured in beacons"
            ret.update({"comment": comt, "result": True})
            assert beacon.absent(beacon_name) == ret



"""
Tests for orchestration events
"""
import concurrent.futures
import functools
import json
import time

import pytest
import salt.utils.jid

pytestmark = [
    pytest.mark.slow_test,
]


def test_state_event(salt_run_cli, salt_cli, salt_minion):
    """
    test to ensure state.event
    runner returns correct data
    """
    # If and when we have tornado 5.0, we can just do io_loop.run_in_executor() ....
    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
        runner_future = executor.submit(
            functools.partial(
                salt_run_cli.run, "state.event", "salt/job/*/new", count=1
            )
        )

        while not runner_future.done():
            ret = salt_cli.run("--static", "test.ping", minion_tgt=salt_minion.id)
            assert ret.exitcode == 0
            assert ret.json is True

        # Wait for the runner command which should now have data to return to us
        exc = runner_future.exception()
        if exc:
            raise exc
        ret = runner_future.result()
        assert ret.exitcode == 0
        # We have to parse the JSON ourselves since we have regular output mixed with JSON output
        data = None
        for line in ret.stdout.splitlines():
            try:
                _, payload = line.strip().split("\t")
                data = json.loads(payload)
                break
            except ValueError:
                continue
        assert isinstance(data, dict)
        assert salt_minion.id in data["minions"]


def test_jid_in_ret_event(
    salt_run_cli, salt_master, salt_minion, event_listener, base_env_state_tree_root_dir
):
    """
    Test to confirm that the ret event for the orchestration contains the
    jid for the jobs spawned.
    """
    test_state_contents = """
    date:
      cmd.run
    """
    test_orch_contents = """
    date_cmd:
      salt.state:
        - tgt: {minion_id}
        - sls: test-state

    ping_minion:
      salt.function:
        - name: test.ping
        - tgt: {minion_id}

    fileserver.file_list:
      salt.runner

    config.values:
      salt.wheel
    """.format(
        minion_id=salt_minion.id
    )
    with pytest.helpers.temp_file(
        "test-state.sls", test_state_contents, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file(
        "test-orch.sls", test_orch_contents, base_env_state_tree_root_dir
    ):
        start_time = time.time()
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.exitcode == 0
        orch_job_data = ret.json
        for step_data in orch_job_data["data"][salt_master.id].values():
            assert "__jid__" in step_data

        expected_event_tag = "salt/run/{}/ret".format(jid)
        event_pattern = (salt_master.id, expected_event_tag)

        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=120
        )
        assert (
            matched_events.found_all_events
        ), "Failed to receive the event with the tag '{}'".format(expected_event_tag)
        for event in matched_events.matches:
            for job_data in event.data["return"]["data"][salt_master.id].values():
                assert "__jid__" in job_data


def test_parallel_orchestrations(
    salt_run_cli, salt_master, salt_minion, event_listener, base_env_state_tree_root_dir
):
    """
    Test to confirm that the parallel state requisite works in orch
    we do this by running 19 test.sleep's of 10 seconds, and a last 10 seconds sleep
    which depends on the previous 19.
    It should take more than 20 seconds and less than 19*10 seconds
    """
    test_orch_contents = """
    {% for count in range(1, 20) %}

    sleep {{ count }}:
        module.run:
            - name: test.sleep
            - length: 10
            - parallel: True

    {% endfor %}

    sleep 21:
        module.run:
            - name: test.sleep
            - length: 10
            - parallel: True
            - require:
                - module: sleep 1
    """
    with pytest.helpers.temp_file(
        "test-orch.sls", test_orch_contents, base_env_state_tree_root_dir
    ):
        start_time = time.time()
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.exitcode == 0
        orch_job_data = ret.json
        for step_data in orch_job_data["data"][salt_master.id].values():
            # we expect each duration to be greater than 10s
            assert step_data["duration"] > 10 * 1000

        # Since we started range(1, 20)(19) sleep state steps, and the last state
        # step requires all of those to have finished before it runs, since we are
        # running in parallel, the duration should be more than 20 seconds, and
        # less than 19*10(190) seconds, less then half actually
        duration = time.time() - start_time
        assert duration > 20
        assert duration < 19 * 10 / 2

        expected_event_tag = "salt/run/{}/ret".format(jid)
        event_pattern = (salt_master.id, expected_event_tag)

        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=120
        )
        assert (
            matched_events.found_all_events
        ), "Failed to receive the event with the tag '{}'".format(expected_event_tag)
        for event in matched_events.matches:
            for job_data in event.data["return"]["data"][salt_master.id].values():
                # we expect each duration to be greater than 10s
                assert job_data["duration"] > 10 * 1000

        # Since we started range(1, 20)(19) sleep state steps, and the last state
        # step requires all of those to have finished before it runs, since we are
        # running in parallel, the duration should be more than 20 seconds, and
        # less than 19*10(190) seconds, less then half actually
        duration = time.time() - start_time
        assert duration > 20
        assert duration < 19 * 10 / 2


def test_orchestration_soft_kill(
    salt_run_cli, salt_master, base_env_state_tree_root_dir
):
    sls_contents = """
    stage_one:
        test.succeed_without_changes

    stage_two:
        test.fail_without_changes
    """
    with pytest.helpers.temp_file(
        "test-orch.sls", sls_contents, base_env_state_tree_root_dir
    ):
        jid = salt.utils.jid.gen_jid(salt_master.config)

        # Without soft kill, the orchestration will fail because stage_two is set to fail
        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.exitcode == 1
        for state_data in ret.json["data"][salt_master.id].values():
            if state_data["__id__"] == "stage_two":
                assert state_data["result"] is False
            else:
                assert state_data["result"] is True

        # With soft kill set, 'stage_two' will not run, thus, the orchestration will not fail
        # and 'stage_two' will not be present in the returned data
        jid = salt.utils.jid.gen_jid(salt_master.config)
        ret = salt_run_cli.run("state.soft_kill", jid, "stage_two")
        assert ret.exitcode == 0
        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.exitcode == 0
        for state_data in ret.json["data"][salt_master.id].values():
            if state_data["__id__"] == "stage_two":
                pytest.fail("'stage_two' was present in the ochestration return data")
            else:
                assert state_data["result"] is True


def test_orchestration_with_pillar_dot_items(
    salt_run_cli, salt_master, base_env_state_tree_root_dir
):
    """
    Test to confirm when using a state file that includes other state file, if
    one of those state files includes pillar related functions that will not
    be pulling from the pillar cache that all the state files are available and
    the file_roots has been preserved.  See issues #48277 and #46986.
    """
    main_sls_contents = """
    include:
      - one
      - two
      - three
    """
    one_sls_contents = """
    {%- set foo = salt['saltutil.runner']('pillar.show_pillar') %}
    placeholder_one:
      test.succeed_without_changes
    """
    two_sls_contents = """
    placeholder_two:
      test.succeed_without_changes
    """
    three_sls_contents = """
    placeholder_three:
      test.succeed_without_changes
    """
    with pytest.helpers.temp_file(
        "test-orch.sls", main_sls_contents, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file(
        "one.sls", one_sls_contents, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file(
        "two.sls", two_sls_contents, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file(
        "three.sls", three_sls_contents, base_env_state_tree_root_dir
    ):
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.exitcode == 0
        for state_data in ret.json["data"][salt_master.id].values():
            # Each state should be successful
            assert state_data["result"] is True


def test_orchestration_onchanges_and_prereq(
    salt_run_cli, salt_master, salt_minion, base_env_state_tree_root_dir, tmp_path
):
    sls_contents = """
    manage_a_file:
      salt.state:
        - tgt: {minion_id}
        - sls:
          - orch-req-test

    do_onchanges:
      salt.function:
        - tgt: {minion_id}
        - name: test.ping
        - onchanges:
          - salt: manage_a_file

    do_prereq:
      salt.function:
        - tgt: {minion_id}
        - name: test.ping
        - prereq:
          - salt: manage_a_file
    """.format(
        minion_id=salt_minion.id
    )

    orch_test_file = tmp_path / "orch-test-file"
    req_sls_contents = """
    {}:
      file.managed:
        - contents: 'Hello World!'
    """.format(
        orch_test_file
    )

    with pytest.helpers.temp_file(
        "test-orch.sls", sls_contents, base_env_state_tree_root_dir
    ), pytest.helpers.temp_file(
        "orch-req-test.sls", req_sls_contents, base_env_state_tree_root_dir
    ):
        jid1 = salt.utils.jid.gen_jid(salt_master.config)

        # Run in test mode, will describe what changes would occur
        ret = salt_run_cli.run(
            "--jid", jid1, "state.orchestrate", "test-orch", test=True
        )
        assert ret.exitcode == 0
        ret1 = ret.json

        # Now run without test mode to actually create the file
        ret = salt_run_cli.run("state.orchestrate", "test-orch")
        assert ret.exitcode == 0

        # Run again in test mode. Since there were no changes, the requisites should not fire.
        jid2 = salt.utils.jid.gen_jid(salt_master.config)
        ret = salt_run_cli.run(
            "--jid", jid2, "state.orchestrate", "test-orch", test=True
        )
        assert ret.exitcode == 0
        ret2 = ret.json

        # The first time through, all three states should have a None result
        for state_data in ret1["data"][salt_master.id].values():
            assert state_data["result"] is None
            if state_data.get("__id__") == "manage_a_file":
                # The file.managed state should have shown changes in the test mode
                # return data.
                assert state_data["changes"]

        # While the second time through, they should all have a True result.
        for state_data in ret2["data"][salt_master.id].values():
            assert state_data["result"] is True
            if state_data.get("__id__") == "manage_a_file":
                # After the file was created, running again in test mode should have
                # shown no changes.
                assert not state_data["changes"]



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.publish as publish
from salt.exceptions import SaltReqTimeoutError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SAuth:
    """
    Mock SAuth class
    """

    def __init__(self, __opts__):
        self.tok = None

    def gen_token(self, tok):
        """
        Mock gen_token method
        """
        self.tok = tok
        return "salt_tok"


class Channel:
    """
    Mock Channel class
    """

    flag = None

    def __init__(self):
        self.tok = None
        self.load = None

    def factory(self, tok):
        """
        Mock factory method
        """
        self.tok = tok
        return Channel()

    def send(self, load):
        """
        Mock send method
        """
        self.load = load
        if self.flag == 1:
            raise SaltReqTimeoutError
        return True

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class PublishTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.publish
    """

    def setup_loader_modules(self):
        return {publish: {}}

    @classmethod
    def setUpClass(cls):
        cls.channel_patcher = patch("salt.transport.client.ReqChannel", Channel())
        cls.channel_patcher.start()

    @classmethod
    def tearDownClass(cls):
        cls.channel_patcher.stop()
        del cls.channel_patcher

    def setUp(self):
        patcher = patch("salt.crypt.SAuth", return_value=SAuth(publish.__opts__))
        patcher.start()
        self.addCleanup(patcher.stop)

    # 'publish' function tests: 1

    def test_publish(self):
        """
        Test if it publish a command from the minion out to other minions.
        """
        self.assertDictEqual(publish.publish("os:Fedora", "publish.salt"), {})

    # 'full_data' function tests: 1

    def test_full_data(self):
        """
        Test if it return the full data about the publication
        """
        self.assertDictEqual(publish.publish("*", "publish.salt"), {})

    # 'runner' function tests: 1

    def test_runner(self):
        """
        Test if it execute a runner on the master and return the data
        from the runner function
        """
        ret = "No access to master. If using salt-call with --local, please remove."
        self.assertEqual(publish.runner("manage.down"), ret)

        mock = MagicMock(return_value=True)
        mock_id = MagicMock(return_value="salt_id")
        with patch.dict(publish.__opts__, {"master_uri": mock, "id": mock_id}):
            Channel.flag = 0
            self.assertTrue(publish.runner("manage.down"))

            Channel.flag = 1
            self.assertEqual(
                publish.runner("manage.down"), "'manage.down' runner publish timed out"
            )



"""
Module for managing logrotate.
"""

import logging
import os

import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError

_LOG = logging.getLogger(__name__)
_DEFAULT_CONF = "/etc/logrotate.conf"


# Define a function alias in order not to shadow built-in's
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return (
            False,
            "The logrotate execution module cannot be loaded: only available "
            "on non-Windows systems.",
        )
    return True


def _convert_if_int(value):
    """
    Convert to an int if necessary.

    :param str value: The value to check/convert.

    :return: The converted or passed value.
    :rtype: bool|int|str
    """
    try:
        value = int(str(value))
    except ValueError:
        pass
    return value


def _parse_conf(conf_file=_DEFAULT_CONF):
    """
    Parse a logrotate configuration file.

    Includes will also be parsed, and their configuration will be stored in the
    return dict, as if they were part of the main config file. A dict of which
    configs came from which includes will be stored in the 'include files' dict
    inside the return dict, for later reference by the user or module.
    """
    ret = {}
    mode = "single"
    multi_names = []
    multi = {}
    prev_comps = None

    with salt.utils.files.fopen(conf_file, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line).strip()
            if not line:
                continue
            if line.startswith("#"):
                continue

            comps = line.split()
            if "{" in line and "}" not in line:
                mode = "multi"
                if len(comps) == 1 and prev_comps:
                    multi_names = prev_comps
                else:
                    multi_names = comps
                    multi_names.pop()
                continue
            if "}" in line:
                mode = "single"
                for multi_name in multi_names:
                    ret[multi_name] = multi
                multi_names = []
                multi = {}
                continue

            if mode == "single":
                key = ret
            else:
                key = multi

            if comps[0] == "include":
                if "include files" not in ret:
                    ret["include files"] = {}
                for include in os.listdir(comps[1]):
                    if include not in ret["include files"]:
                        ret["include files"][include] = []

                    include_path = os.path.join(comps[1], include)
                    include_conf = _parse_conf(include_path)

                    for file_key in include_conf:
                        ret[file_key] = include_conf[file_key]
                        ret["include files"][include].append(file_key)

            prev_comps = comps
            if len(comps) > 2:
                key[comps[0]] = " ".join(comps[1:])
            elif len(comps) > 1:
                key[comps[0]] = _convert_if_int(comps[1])
            else:
                key[comps[0]] = True
    return ret


def show_conf(conf_file=_DEFAULT_CONF):
    """
    Show parsed configuration

    :param str conf_file: The logrotate configuration file.

    :return: The parsed configuration.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' logrotate.show_conf
    """
    return _parse_conf(conf_file)


def get(key, value=None, conf_file=_DEFAULT_CONF):
    """
    Get the value for a specific configuration line.

    :param str key: The command or stanza block to configure.
    :param str value: The command value or command of the block specified by the key parameter.
    :param str conf_file: The logrotate configuration file.

    :return: The value for a specific configuration line.
    :rtype: bool|int|str

    CLI Example:

    .. code-block:: bash

        salt '*' logrotate.get rotate

        salt '*' logrotate.get /var/log/wtmp rotate /etc/logrotate.conf
    """
    current_conf = _parse_conf(conf_file)
    stanza = current_conf.get(key, False)

    if value:
        if stanza:
            return stanza.get(value, False)
        _LOG.warning("Block '%s' not present or empty.", key)
    return stanza


def set_(key, value, setting=None, conf_file=_DEFAULT_CONF):
    """
    Set a new value for a specific configuration line.

    :param str key: The command or block to configure.
    :param str value: The command value or command of the block specified by the key parameter.
    :param str setting: The command value for the command specified by the value parameter.
    :param str conf_file: The logrotate configuration file.

    :return: A boolean representing whether all changes succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' logrotate.set rotate 2

    Can also be used to set a single value inside a multiline configuration
    block. For instance, to change rotate in the following block:

    .. code-block:: text

        /var/log/wtmp {
            monthly
            create 0664 root root
            rotate 1
        }

    Use the following command:

    .. code-block:: bash

        salt '*' logrotate.set /var/log/wtmp rotate 2

    This module also has the ability to scan files inside an include directory,
    and make changes in the appropriate file.
    """
    conf = _parse_conf(conf_file)
    for include in conf["include files"]:
        if key in conf["include files"][include]:
            conf_file = os.path.join(conf["include"], include)

    new_line = ""
    kwargs = {
        "flags": 8,
        "backup": False,
        "path": conf_file,
        "pattern": "^{}.*".format(key),
        "show_changes": False,
    }

    if setting is None:
        current_value = conf.get(key, False)

        if isinstance(current_value, dict):
            raise SaltInvocationError(
                "Error: {} includes a dict, and a specific setting inside the "
                "dict was not declared".format(key)
            )

        if value == current_value:
            _LOG.debug("Command '%s' already has: %s", key, value)
            return True

        # This is the new config line that will be set
        if value is True:
            new_line = key
        elif value:
            new_line = "{} {}".format(key, value)

        kwargs.update({"prepend_if_not_found": True})
    else:
        stanza = conf.get(key, dict())

        if stanza and not isinstance(stanza, dict):
            error_msg = (
                "Error: A setting for a dict was declared, but the "
                "configuration line given is not a dict"
            )
            raise SaltInvocationError(error_msg)

        if setting == stanza.get(value, False):
            _LOG.debug("Command '%s' already has: %s", value, setting)
            return True

        # We're going to be rewriting an entire stanza
        if setting:
            stanza[value] = setting
        else:
            del stanza[value]

        new_line = _dict_to_stanza(key, stanza)

        kwargs.update(
            {
                "pattern": "^{0}.*?{{.*?}}".format(key),
                "flags": 24,
                "append_if_not_found": True,
            }
        )

    kwargs.update({"repl": new_line})
    _LOG.debug("Setting file '%s' line: %s", conf_file, new_line)

    return __salt__["file.replace"](**kwargs)


def _dict_to_stanza(key, stanza):
    """
    Convert a dict to a multi-line stanza
    """
    ret = ""
    for skey in stanza:
        if stanza[skey] is True:
            stanza[skey] = ""
        ret += "    {} {}\n".format(skey, stanza[skey])
    return "{0} {{\n{1}}}".format(key, ret)



"""
Test the elasticsearch returner
"""
import salt.returners.elasticsearch_return as elasticsearch_return
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ElasticSearchReturnerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test the functions in the elasticsearch returner
    """

    def setup_loader_modules(self):
        return {elasticsearch_return: {}}

    def test__virtual_no_elasticsearch(self):
        """
        Test __virtual__ function when elasticsearch is not installed
        and the elasticsearch module is not available
        """
        result = elasticsearch_return.__virtual__()
        expected = (
            False,
            "Elasticsearch module not availble.  Check that the elasticsearch library"
            " is installed.",
        )
        self.assertEqual(expected, result)

    def test__virtual_with_elasticsearch(self):
        """
        Test __virtual__ function when elasticsearch
        and the elasticsearch module is not available
        """
        with patch.dict(
            elasticsearch_return.__salt__, {"elasticsearch.index_exists": MagicMock()}
        ):
            result = elasticsearch_return.__virtual__()
            expected = "elasticsearch"
            self.assertEqual(expected, result)



"""
Package support for Solaris

.. important::
    If you feel that Salt should be using this module to manage packages on a
    minion, and it is using a different module (or gives an error similar to
    *'pkg.install' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import copy
import logging
import os

import salt.utils.data
import salt.utils.files
import salt.utils.functools
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, MinionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Set the virtual pkg module if the os is Solaris
    """
    if (
        __grains__["os_family"] == "Solaris"
        and float(__grains__["kernelrelease"]) <= 5.10
    ):
        return __virtualname__
    return (
        False,
        "The solarispkg execution module failed to load: only available "
        "on Solaris <= 10.",
    )


def _write_adminfile(kwargs):
    """
    Create a temporary adminfile based on the keyword arguments passed to
    pkg.install.
    """
    # Set the adminfile default variables
    email = kwargs.get("email", "")
    instance = kwargs.get("instance", "quit")
    partial = kwargs.get("partial", "nocheck")
    runlevel = kwargs.get("runlevel", "nocheck")
    idepend = kwargs.get("idepend", "nocheck")
    rdepend = kwargs.get("rdepend", "nocheck")
    space = kwargs.get("space", "nocheck")
    setuid = kwargs.get("setuid", "nocheck")
    conflict = kwargs.get("conflict", "nocheck")
    action = kwargs.get("action", "nocheck")
    basedir = kwargs.get("basedir", "default")

    # Make tempfile to hold the adminfile contents.
    adminfile = salt.utils.files.mkstemp(prefix="salt-")

    def _write_line(fp_, line):
        fp_.write(salt.utils.stringutils.to_str(line))

    with salt.utils.files.fopen(adminfile, "w") as fp_:
        _write_line(fp_, "email={}\n".format(email))
        _write_line(fp_, "instance={}\n".format(instance))
        _write_line(fp_, "partial={}\n".format(partial))
        _write_line(fp_, "runlevel={}\n".format(runlevel))
        _write_line(fp_, "idepend={}\n".format(idepend))
        _write_line(fp_, "rdepend={}\n".format(rdepend))
        _write_line(fp_, "space={}\n".format(space))
        _write_line(fp_, "setuid={}\n".format(setuid))
        _write_line(fp_, "conflict={}\n".format(conflict))
        _write_line(fp_, "action={}\n".format(action))
        _write_line(fp_, "basedir={}\n".format(basedir))

    return adminfile


def _list_pkgs_from_context(versions_as_list):
    """
    Use pkg list from __context__
    """
    if versions_as_list:
        return __context__["pkg.list_pkgs"]
    else:
        ret = copy.deepcopy(__context__["pkg.list_pkgs"])
        __salt__["pkg_resource.stringify"](ret)
        return ret


def list_pkgs(versions_as_list=False, **kwargs):
    """
    List the packages currently installed as a dict:

    .. code-block:: python

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    """
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any(
        [salt.utils.data.is_true(kwargs.get(x)) for x in ("removed", "purge_desired")]
    ):
        return {}

    if "pkg.list_pkgs" in __context__ and kwargs.get("use_context", True):
        return _list_pkgs_from_context(versions_as_list)

    ret = {}
    cmd = "/usr/bin/pkginfo -x"

    # Package information returned two lines per package. On even-offset
    # lines, the package name is in the first column. On odd-offset lines, the
    # package version is in the second column.
    lines = __salt__["cmd.run"](
        cmd, output_loglevel="trace", python_shell=False
    ).splitlines()
    for index, line in enumerate(lines):
        if index % 2 == 0:
            name = line.split()[0].strip()
        if index % 2 == 1:
            version_num = line.split()[1].strip()
            __salt__["pkg_resource.add_pkg"](ret, name, version_num)

    __salt__["pkg_resource.sort_pkglist"](ret)
    __context__["pkg.list_pkgs"] = copy.deepcopy(ret)
    if not versions_as_list:
        __salt__["pkg_resource.stringify"](ret)
    return ret


def latest_version(*names, **kwargs):
    """
    Return the latest version of the named package available for upgrade or
    installation. If more than one package name is specified, a dict of
    name/version pairs is returned.

    If the latest version of a given package is already installed, an empty
    string will be returned for that package.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
        salt '*' pkg.latest_version <package1> <package2> <package3> ...

    NOTE: As package repositories are not presently supported for Solaris
    pkgadd, this function will always return an empty string for a given
    package.
    """
    kwargs.pop("refresh", True)

    ret = {}
    if not names:
        return ""
    for name in names:
        ret[name] = ""

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret


# available_version is being deprecated
available_version = salt.utils.functools.alias_function(
    latest_version, "available_version"
)


def upgrade_available(name, **kwargs):
    """
    Check whether or not an upgrade is available for a given package

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>
    """
    return latest_version(name) != ""


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    """
    return __salt__["pkg_resource.version"](*names, **kwargs)


def install(name=None, sources=None, saltenv="base", **kwargs):
    """
    Install the passed package. Can install packages from the following
    sources:

    * Locally (package already exists on the minion
    * HTTP/HTTPS server
    * FTP server
    * Salt master

    Returns a dict containing the new package names and versions:

    .. code-block:: python

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Examples:

    .. code-block:: bash

        # Installing a data stream pkg that already exists on the minion

        salt '*' pkg.install sources='[{"<pkg name>": "/dir/on/minion/<pkg filename>"}]'
        salt '*' pkg.install sources='[{"SMClgcc346": "/var/spool/pkg/gcc-3.4.6-sol10-sparc-local.pkg"}]'

        # Installing a data stream pkg that exists on the salt master

        salt '*' pkg.install sources='[{"<pkg name>": "salt://pkgs/<pkg filename>"}]'
        salt '*' pkg.install sources='[{"SMClgcc346": "salt://pkgs/gcc-3.4.6-sol10-sparc-local.pkg"}]'

    CLI Example:

    .. code-block:: bash

        # Installing a data stream pkg that exists on a HTTP server
        salt '*' pkg.install sources='[{"<pkg name>": "http://packages.server.com/<pkg filename>"}]'
        salt '*' pkg.install sources='[{"SMClgcc346": "http://packages.server.com/gcc-3.4.6-sol10-sparc-local.pkg"}]'

    If working with solaris zones and you want to install a package only in the
    global zone you can pass 'current_zone_only=True' to salt to have the
    package only installed in the global zone. (Behind the scenes this is
    passing '-G' to the pkgadd command.) Solaris default when installing a
    package in the global zone is to install it in all zones. This overrides
    that and installs the package only in the global.

    CLI Example:

    .. code-block:: bash

        # Installing a data stream package only in the global zone:
        salt 'global_zone' pkg.install sources='[{"SMClgcc346": "/var/spool/pkg/gcc-3.4.6-sol10-sparc-local.pkg"}]' current_zone_only=True

    By default salt automatically provides an adminfile, to automate package
    installation, with these options set::

        email=
        instance=quit
        partial=nocheck
        runlevel=nocheck
        idepend=nocheck
        rdepend=nocheck
        space=nocheck
        setuid=nocheck
        conflict=nocheck
        action=nocheck
        basedir=default

    You can override any of these options in two ways. First you can optionally
    pass any of the options as a kwarg to the module/state to override the
    default value or you can optionally pass the 'admin_source' option
    providing your own adminfile to the minions.

    Note: You can find all of the possible options to provide to the adminfile
    by reading the admin man page:

    .. code-block:: bash

        man -s 4 admin

    CLI Example:

    .. code-block:: bash

        # Overriding the 'instance' adminfile option when calling the module directly
        salt '*' pkg.install sources='[{"<pkg name>": "salt://pkgs/<pkg filename>"}]' instance="overwrite"

    SLS Example:

    .. code-block:: yaml

        # Overriding the 'instance' adminfile option when used in a state

        SMClgcc346:
          pkg.installed:
            - sources:
              - SMClgcc346: salt://srv/salt/pkgs/gcc-3.4.6-sol10-sparc-local.pkg
            - instance: overwrite

    .. note::
        The ID declaration is ignored, as the package name is read from the
        ``sources`` parameter.

    CLI Example:

    .. code-block:: bash

        # Providing your own adminfile when calling the module directly

        salt '*' pkg.install sources='[{"<pkg name>": "salt://pkgs/<pkg filename>"}]' admin_source='salt://pkgs/<adminfile filename>'

        # Providing your own adminfile when using states

        <pkg name>:
          pkg.installed:
            - sources:
              - <pkg name>: salt://pkgs/<pkg filename>
            - admin_source: salt://pkgs/<adminfile filename>

    .. note::
        The ID declaration is ignored, as the package name is read from the
        ``sources`` parameter.
    """
    if salt.utils.data.is_true(kwargs.get("refresh")):
        log.warning("'refresh' argument not implemented for solarispkg module")

    # pkgs is not supported, but must be passed here for API compatibility
    pkgs = kwargs.pop("pkgs", None)
    try:
        pkg_params, pkg_type = __salt__["pkg_resource.parse_targets"](
            name, pkgs, sources, **kwargs
        )
    except MinionError as exc:
        raise CommandExecutionError(exc)

    if pkg_params is None or len(pkg_params) == 0:
        return {}

    if not sources:
        log.error('"sources" param required for solaris pkg_add installs')
        return {}

    try:
        if "admin_source" in kwargs:
            adminfile = __salt__["cp.cache_file"](kwargs["admin_source"], saltenv)
        else:
            adminfile = _write_adminfile(kwargs)

        old = list_pkgs()
        cmd_prefix = ["/usr/sbin/pkgadd", "-n", "-a", adminfile]

        # Only makes sense in a global zone but works fine in non-globals.
        if kwargs.get("current_zone_only") in (True, "True"):
            cmd_prefix += "-G "

        errors = []
        for pkg in pkg_params:
            cmd = cmd_prefix + ["-d", pkg, "all"]
            # Install the package{s}
            out = __salt__["cmd.run_all"](
                cmd, output_loglevel="trace", python_shell=False
            )

            if out["retcode"] != 0 and out["stderr"]:
                errors.append(out["stderr"])

        __context__.pop("pkg.list_pkgs", None)
        new = list_pkgs()
        ret = salt.utils.data.compare_dicts(old, new)

        if errors:
            raise CommandExecutionError(
                "Problem encountered installing package(s)",
                info={"errors": errors, "changes": ret},
            )
    finally:
        # Remove the temp adminfile
        if "admin_source" not in kwargs:
            try:
                os.remove(adminfile)
            except (NameError, OSError):
                pass

    return ret


def remove(name=None, pkgs=None, saltenv="base", **kwargs):
    """
    Remove packages with pkgrm

    name
        The name of the package to be deleted

    By default salt automatically provides an adminfile, to automate package
    removal, with these options set::

        email=
        instance=quit
        partial=nocheck
        runlevel=nocheck
        idepend=nocheck
        rdepend=nocheck
        space=nocheck
        setuid=nocheck
        conflict=nocheck
        action=nocheck
        basedir=default

    You can override any of these options in two ways. First you can optionally
    pass any of the options as a kwarg to the module/state to override the
    default value or you can optionally pass the 'admin_source' option
    providing your own adminfile to the minions.

    Note: You can find all of the possible options to provide to the adminfile
    by reading the admin man page:

    .. code-block:: bash

        man -s 4 admin


    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <package name>
        salt '*' pkg.remove SUNWgit
        salt '*' pkg.remove <package1>,<package2>,<package3>
        salt '*' pkg.remove pkgs='["foo", "bar"]'
    """
    try:
        pkg_params = __salt__["pkg_resource.parse_targets"](name, pkgs)[0]
    except MinionError as exc:
        raise CommandExecutionError(exc)

    old = list_pkgs()
    targets = [x for x in pkg_params if x in old]
    if not targets:
        return {}

    try:
        if "admin_source" in kwargs:
            adminfile = __salt__["cp.cache_file"](kwargs["admin_source"], saltenv)
        else:
            # Make tempfile to hold the adminfile contents.
            adminfile = _write_adminfile(kwargs)

        # Remove the package
        cmd = ["/usr/sbin/pkgrm", "-n", "-a", adminfile] + targets
        out = __salt__["cmd.run_all"](cmd, python_shell=False, output_loglevel="trace")

        if out["retcode"] != 0 and out["stderr"]:
            errors = [out["stderr"]]
        else:
            errors = []

        __context__.pop("pkg.list_pkgs", None)
        new = list_pkgs()
        ret = salt.utils.data.compare_dicts(old, new)

        if errors:
            raise CommandExecutionError(
                "Problem encountered removing package(s)",
                info={"errors": errors, "changes": ret},
            )
    finally:
        # Remove the temp adminfile
        if "admin_source" not in kwargs:
            try:
                os.remove(adminfile)
            except (NameError, OSError):
                pass

    return ret


def purge(name=None, pkgs=None, **kwargs):
    """
    Package purges are not supported, this function is identical to
    ``remove()``.

    name
        The name of the package to be deleted


    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.purge <package name>
        salt '*' pkg.purge <package1>,<package2>,<package3>
        salt '*' pkg.purge pkgs='["foo", "bar"]'
    """
    return remove(name=name, pkgs=pkgs, **kwargs)
